'''
Test clingo's Application class.
'''

from typing import Any, Callable, List, Sequence, Tuple
from unittest import TestCase
from tempfile import NamedTemporaryFile
from multiprocessing import Process, Queue
import os
import re

from .util import _MCB
from ..core import MessageCode
from ..application import Application, ApplicationOptions, Flag, clingo_main

class TestApp(Application):
    '''
    Test application covering most of the Application related API.

    Note that I did not find a nice way to test model printing.
    '''
    _queue: Queue
    program_name = 'test'
    version = '1.2.3'
    message_limit = 17

    def __init__(self, queue: Queue):
        self._queue = queue
        self._flag = Flag()

    def _parse_test(self, value):
        self._queue.put(('parse', value))
        return True

    def register_options(self, options: ApplicationOptions) -> None:
        self._queue.put('register')
        group = 'Clingo.Test'
        options.add(group, 'test', 'test description', self._parse_test)
        options.add_flag(group, 'flag', 'test description', self._flag)

    def validate_options(self) -> bool:
        self._queue.put('validate')
        self._queue.put(('flag', self._flag.flag))
        return True

    def logger(self, code: MessageCode, message: str) -> None:
        self._queue.put((code, re.sub('^.*:(?=[0-9]+:)', '', message)))

    def main(self, control, files):
        self._queue.put('main')
        for file_ in files:
            control.load(file_)
        control.ground([("base", [])])
        mcb = _MCB()
        control.solve(on_model=mcb.on_model)
        self._queue.put(('models', [ [ str(sym) for sym in model ] for model in mcb.models ]))

def _run_process(app: Callable[[Queue], Application], program: str, queue: Queue, args: Sequence[str]) -> None:
    '''
    Run clingo application with given program and intercept results.
    '''
    with NamedTemporaryFile(mode='wt', delete=False) as fp:
        name = fp.name
        fp.write(program)
    try:
        # Note: The multiprocess module does not allow for intercepting the
        # output. Thus, the output is simply disabled and we use the Queue
        # class to communicate results.
        ret = clingo_main(app(queue), (name, '--outf=3') + tuple(args))
        queue.put(int(ret))
        queue.close()
    finally:
        os.unlink(name)

AppResult = Tuple[int, List[Any]]

def run_app(app: Callable[[Queue], Application], program: str, *args: Sequence[str]) -> AppResult:
    '''
    Run clingo application in subprocess via multiprocessing module.
    '''
    q: Queue
    q = Queue()
    p = Process(target=_run_process, args=(app, program, q, tuple(args)))

    p.start()
    seq: List[Any]
    seq, ret = [], -1
    while True:
        ret = q.get()
        if isinstance(ret, int):
            status = ret
            break
        seq.append(ret)
    p.join()
    q.close()

    return status, seq

class TestApplication(TestCase):
    '''
    Tests for clingo's application class.
    '''
    def test_app(self):
        '''
        Test application.
        '''
        ret, seq = run_app(TestApp, "1 {a; b; c(1/0)}.", "0", '--test=x', '--flag')
        self.assertEqual(ret, 30)
        self.assertEqual(seq, [
            'register',
            ('parse', 'x'),
            'validate',
            ('flag', True),
            'main',
            (MessageCode.OperationUndefined, '1:12-15: info: operation undefined:\n  (1/0)\n'),
            ('models', [['a'], ['a', 'b'], ['b']])])



'''
Tests for the ast module.
'''
from unittest import TestCase
from typing import cast
from textwrap import dedent
from collections.abc import Sequence
from copy import copy, deepcopy

from .. import ast
from ..ast import (AST, ASTSequence, Id, Location, Program, ProgramBuilder, Position, StrSequence,
                   SymbolicTerm, TheoryUnparsedTermElement,
                   parse_string)
from ..symbol import Function, Number
from ..control import Control

class VariableRenamer(ast.Transformer):
    '''
    Add an underscore to all names of variables in an ast.
    '''
    def visit_Variable(self, node): #pylint: disable=no-self-use,invalid-name
        '''
        Rename a variable.
        '''
        return ast.Variable(node.location, '_' + node.name)

class TestAST(TestCase):
    '''
    Tests for the ast module.
    '''
    def _deepcopy(self, node: AST) -> AST:
        '''
        This functions tests manual deep copying of ast nodes using all the
        constructor functions in the ast module.

        Since this functions visits all possible AST nodes, further
        functionality of the AST is tested here.
        '''
        cons_name = str(node.ast_type).split('.')[1]
        cons = getattr(ast, cons_name)
        args = cast(dict, dict(node.items()))

        def to_list(val):
            return list(val) if isinstance(val, Sequence) else val

        items = [ to_list(val) for key, val in node.items() ]
        zipped = [ to_list(val) for key, val in zip(node.keys(), node.values()) ]
        self.assertEqual(items, zipped)

        for key in node.child_keys:
            if isinstance(args[key], Sequence):
                args[key] = [ self._deepcopy(cast(AST, child)) for child in args[key] ]
            elif isinstance(args[key], AST):
                args[key] = self._deepcopy(cast(AST, args[key]))

        cpy = cons(**args)
        self.assertEqual(cpy, node)

        for key in node.keys():
            setattr(cpy, key, getattr(cpy, key))

        cpz = copy(cpy)
        for key in node.keys():
            setattr(cpy, key, getattr(cpz, key))

        self.assertEqual(cpy, node)

        return cpy

    def _str(self, s, alt=None):
        prg = []
        parse_string(s, prg.append)
        cpy = copy(deepcopy(self._deepcopy(prg[-1])))
        self.assertEqual(str(cpy), s if alt is None else alt)

        prg = []
        parse_string(str(cpy), prg.append)
        self.assertEqual(str(prg[-1]), s if alt is None else alt)

        try:
            ctl = Control()
            with ProgramBuilder(ctl) as bld:
                bld.add(prg[-1])
        except RuntimeError as e:
            msg = e.args[0]
            if ("error: python support not available" not in msg and
                "error: lua support not available" not in msg):
                raise RuntimeError from e

    def test_terms(self):
        '''
        Test terms.
        '''
        self._str("a.")
        self._str("-a.")
        self._str("a(X).")
        self._str("a(-X).")
        self._str("a(|X|).")
        self._str("a(~X).")
        self._str("a((X^Y)).")
        self._str("a((X?Y)).")
        self._str("a((X&Y)).")
        self._str("a((X+Y)).")
        self._str("a((X-Y)).")
        self._str("a((X*Y)).")
        self._str("a((X/Y)).")
        self._str("a((X\\Y)).")
        self._str("a((X**Y)).")
        self._str("a((X..Y)).")
        self._str("-a(f).")
        self._str("-a(-f).")
        self._str("-a(f(X)).")
        self._str("-a(f(X,Y)).")
        self._str("-a(()).")
        self._str("-a((a,)).")
        self._str("-a((a,b)).")
        self._str("-a(@f(a,b)).")
        self._str("-a(@f).")
        self._str("-a(a;b;c).")
        self._str("-a((a;b;c)).")
        self._str("-a(f(a);f(b);f(c)).")

    def test_theory_terms(self):
        '''
        Test theory terms.
        '''
        self._str("&a { 1 }.")
        self._str("&a { (- 1) }.")
        self._str("&a { X }.")
        self._str("&a { () }.")
        self._str("&a { (1,) }.")
        self._str("&a { (1,2) }.")
        self._str("&a { [] }.")
        self._str("&a { [1] }.")
        self._str("&a { [1,2] }.")
        self._str("&a { {} }.")
        self._str("&a { {1} }.")
        self._str("&a { {1,2} }.")
        self._str("&a { f }.")
        self._str("&a { f(X) }.")
        self._str("&a { f(X,Y) }.")
        self._str("&a { (+ a + - * b + c) }.")

    def test_literals(self):
        '''
        Test literals.
        '''
        self._str("a.")
        self._str("not a.")
        self._str("not not a.")
        self._str("1 < 2.")
        self._str("1 <= 2.")
        self._str("1 > 2.")
        self._str("1 >= 2.")
        self._str("1 = 2.")
        self._str("1 != 2.")
        self._str("#false.")
        self._str("#true.")

    def test_head_literals(self):
        '''
        Test head literals.
        '''
        self._str("{ }.")
        self._str("{ } < 2.", "2 > { }.")
        self._str("1 < { }.")
        self._str("1 < { } < 2.")
        self._str("{ b }.")
        self._str("{ a; b }.")
        self._str("{ a; b: c, d }.")
        self._str("#count { }.")
        self._str("#count { } < 2.", "2 > #count { }.")
        self._str("1 < #count { }.")
        self._str("1 < #count { } < 2.")
        self._str("#count { b: a }.")
        self._str("#count { b,c: a }.")
        self._str("#count { a: a; b: c }.")
        self._str("#count { a: d; b: x: c, d }.")
        self._str("#min { }.")
        self._str("#max { }.")
        self._str("#sum { }.")
        self._str("#sum+ { }.")
        self._str("a; b.")
        self._str("a; b: c.")
        self._str("a; b: c, d.")
        self._str("&a { }.")
        self._str("&a { 1 }.")
        self._str("&a { 1; 2 }.")
        self._str("&a { 1,2 }.")
        self._str("&a { 1,2: a }.")
        self._str("&a { 1,2: a, b }.")
        self._str("&a { } != x.")
        self._str("&a(x) { }.")

    def test_body_literals(self):
        '''
        Test body literals.
        '''
        self._str("a :- { }.")
        self._str("a :- not { }.")
        self._str("a :- not not { }.")
        self._str("a :- { } < 2.", "a :- 2 > { }.")
        self._str("a :- 1 < { }.")
        self._str("a :- 1 < { } < 2.")
        self._str("a :- { b }.")
        self._str("a :- { a; b }.")
        self._str("a :- { a; b: c, d }.")
        self._str("a :- #count { }.")
        self._str("a :- not #count { }.")
        self._str("a :- not not #count { }.")
        self._str("a :- #count { } < 2.", "a :- 2 > #count { }.")
        self._str("a :- 1 < #count { }.")
        self._str("a :- 1 < #count { } < 2.")
        self._str("a :- #count { b }.")
        self._str("a :- #count { b,c }.")
        self._str("a :- #count { a; b }.")
        self._str("a :- #count { a; b: c, d }.")
        self._str("a :- #min { }.")
        self._str("a :- #max { }.")
        self._str("a :- #sum { }.")
        self._str("a :- #sum+ { }.")
        self._str("a :- a; b.")
        self._str("a :- a; b: c.")
        self._str("a :- a; b: c, d.")
        self._str("a :- &a { }.")
        self._str("a :- &a { 1 }.")
        self._str("a :- &a { 1; 2 }.")
        self._str("a :- &a { 1,2 }.")
        self._str("a :- &a { 1,2: a }.")
        self._str("a :- &a { 1,2: a, b }.")
        self._str("a :- &a { } != x.")
        self._str("a :- &a(x) { }.")
        self._str("a :- a.")
        self._str("a :- not a.")
        self._str("a :- not not a.")
        self._str("a :- 1 < 2.")
        self._str("a :- 1 <= 2.")
        self._str("a :- 1 > 2.")
        self._str("a :- 1 >= 2.")
        self._str("a :- 1 = 2.")
        self._str("a :- 1 != 2.")
        self._str("a :- #false.")
        self._str("a :- #true.")

    def test_statements(self):
        '''
        Test statements.
        '''
        self._str("a.")
        self._str("#false.")
        self._str("#false :- a.")
        self._str("a :- a; b.")
        self._str("#const x = 10.")
        self._str("#const x = 10. [override]")
        self._str("#show p/1.")
        self._str("#show -p/1.")
        self._str("#defined p/1.")
        self._str("#defined -p/1.")
        self._str("#show x.")
        self._str("#show x : y; z.")
        self._str(":~ . [1@0]")
        self._str(":~ b; c. [1@2,s,t]")
        self._str("#script (lua)\ncode\n#end.")
        self._str("#script (python)\ncode\n#end.")
        self._str("#program x(y, z).")
        self._str("#program x.")
        self._str("#external a. [X]")
        self._str("#external a : b; c. [false]")
        self._str("#edge (1,2).")
        self._str("#edge (1,2) : x; y.")
        self._str("#heuristic a. [b@p,m]")
        self._str("#heuristic a : b; c. [b@p,m]")
        self._str("#project a.")
        self._str("#project a : b; c.")
        self._str("#project -a/0.")
        self._str("#project a/0.")
        self._str("#theory x {\n}.")
        self._str(dedent("""\
                         #theory x {
                           t {
                             + : 0, unary;
                             - : 1, binary, left;
                             * : 2, binary, right
                           };
                           &a/0: t, head;
                           &b/0: t, body;
                           &c/0: t, directive;
                           &d/0: t, { }, t, any;
                           &e/0: t, { =, !=, + }, t, any
                         }."""))

    def test_csp(self):
        '''
        Test CSP constructs.
        '''
        self._str("#false :- #disjoint { }.")
        self._str("#false :- #disjoint { a:2$*$x; b:3$*$y }.")
        self._str("#false :- #disjoint { a:2$*$x: p; b:3$*$y: q, r }.")
        self._str("0 $<= 1$*$(x;y;z) $<= 3.")
        self._str("1$*$x$+1$*$y$+-1$*$z $<= -2.")

    def test_compare(self):
        """
        Test comparison and hashing.
        """
        pos = Position("<string>", 1, 1)
        loc = Location(pos, pos)
        x = Id(location=loc, name="x")
        y = Id(location=Location(pos, Position("<string>", 1, 2)), name="x")
        z = Id(location=loc, name="z")
        self.assertEqual(x, y)
        self.assertEqual(x, x)
        self.assertNotEqual(x, z)
        self.assertEqual(hash(x), hash(x))
        self.assertEqual(hash(x), hash(y))
        self.assertNotEqual(hash(x), hash(z))
        self.assertLess(x, z)
        self.assertNotEqual(x, z)
        self.assertGreater(z, x)
        self.assertLessEqual(y, x)
        self.assertLessEqual(x, y)
        self.assertGreaterEqual(y, x)
        self.assertGreaterEqual(x, y)

    def test_ast_sequence(self):
        """
        Test ast sequences.
        """
        pos = Position("<string>", 1, 1)
        loc = Location(pos, pos)
        lst = [Id(loc, "x"), Id(loc, "y"), Id(loc, "z")]
        prg = Program(loc, "p", lst)
        seq = prg.parameters
        self.assertIsInstance(seq, ASTSequence)
        self.assertEqual(len(seq), 3)
        self.assertEqual(list(seq), lst)
        self.assertEqual(seq[0], lst[0])
        seq.insert(0, Id(loc, "i"))
        self.assertEqual(list(seq), [Id(loc, "i")] + lst)
        seq.insert(0, seq[3])
        self.assertEqual(list(seq), [Id(loc, "z"), Id(loc, "i")] + lst)
        del seq[2]
        self.assertEqual(list(seq), [Id(loc, "z"), Id(loc, "i")] + lst[1:])

    def test_str_sequence(self):
        """
        Test ast sequences.
        """
        pos = Position("<string>", 1, 1)
        loc = Location(pos, pos)

        lst = ["x", "y", "z"]
        sym = SymbolicTerm(loc, Function("a", [Number(1)]))
        tue = TheoryUnparsedTermElement(lst, sym)
        seq = tue.operators
        self.assertIsInstance(seq, StrSequence)
        self.assertEqual(len(seq), 3)
        self.assertEqual(list(seq), lst)
        self.assertEqual(seq[0], lst[0])
        seq.insert(0, "i")
        self.assertEqual(list(seq), ["i"] + lst)
        seq.insert(0, seq[3])
        self.assertEqual(list(seq), ["z", "i"] + lst)
        del seq[2]
        self.assertEqual(list(seq), ["z", "i"] + lst[1:])

    def test_unpool(self):
        """
        Test unpooling.
        """
        prg = []
        parse_string(":- a(1;2): a(3;4).", prg.append)
        lit = prg[-1].body[0]
        def unpool(other=True, condition=True):
            return [ str(x) for x in lit.unpool(other, condition) ]

        self.assertEqual(unpool(),
                         ['a(1): a(3)',
                          'a(1): a(4)',
                          'a(2): a(3)',
                          'a(2): a(4)'])
        self.assertEqual(unpool(other=False),
                         ['a(1;2): a(3)',
                          'a(1;2): a(4)'])
        self.assertEqual(unpool(condition=False),
                         ['a(1): a(3;4)',
                          'a(2): a(3;4)'])
        self.assertEqual(unpool(other=False, condition=False),
                         ['a(1;2): a(3;4)'])

    def test_transformer(self):
        '''
        Test the transformer class.
        '''
        prg = []
        vrt = VariableRenamer()
        parse_string('p(X) :- q(X).', lambda stm: prg.append(str(vrt(stm))))
        self.assertEqual(prg[-1], 'p(_X) :- q(_X).')

    def test_repr(self):
        '''
        Test the string representation of AST nodes.
        '''
        # pylint: disable=eval-used
        stms = []
        prg = dedent('''\
            a :- &sum(body,second) { foo; (1 - 3) } > (4 * 17).
            &theory { (X * a ** stuff): dom(X); (1 - 3) }.
            &diff { (foo - bar) } <= 42.
            1 #max { X : a } :- a, X = #count { a }.
            ''')
        parse_string(prg, stms.append)
        self.assertEqual(stms, eval(repr(stms)))



'''
Functions and classes to observe or add ground statements.

Examples
--------
The following example shows how to add a fact to a program via the backend and
observe the corresponding rule passed to the backend:

    >>> from clingo.symbol import Function
    >>> from clingo.control import Control
    >>>
    >>> class Observer:
    ...     def rule(self, choice, head, body):
    ...         print("rule:", choice, head, body)
    ...
    >>> ctl = Control()
    >>> ctl.register_observer(Observer())
    >>>
    >>> sym_a = Function("a")
    >>> with ctl.backend() as backend:
    ...     atm_a = backend.add_atom(sym_a)
    ...     backend.add_rule([atm_a])
    ...
    rule: False [1] []
    >>> ctl.symbolic_atoms[sym_a].is_fact
    True
    >>>
    >>> print(ctl.solve(on_model=print))
    a
    SAT
'''

from typing import ContextManager, Sequence, Optional, Tuple
from enum import Enum
from abc import ABCMeta

from ._internal import _c_call, _cb_error_handler, _ffi, _handle_error, _lib, _to_str
from .core import TruthValue
from .symbol import Symbol

__all__ = [ 'Backend', 'HeuristicType', 'Observer' ]

class HeuristicType(Enum):
    '''
    Enumeration of the different heuristic types.
    '''
    Factor = _lib.clingo_heuristic_type_factor
    '''
    Heuristic modification to set the decaying factor of an atom.
    '''
    False_ = _lib.clingo_heuristic_type_false
    '''
    Heuristic modification to make an atom false.
    '''
    Init = _lib.clingo_heuristic_type_init
    '''
    Heuristic modification to set the inital score of an atom.
    '''
    Level = _lib.clingo_heuristic_type_level
    '''
    Heuristic modification to set the level of an atom.
    '''
    Sign = _lib.clingo_heuristic_type_sign
    '''
    Heuristic modification to set the sign of an atom.
    '''
    True_ = _lib.clingo_heuristic_type_true
    '''
    Heuristic modification to make an atom true.
    '''

class Observer(metaclass=ABCMeta):
    '''
    Interface that has to be implemented to inspect rules produced during
    grounding.

    See Also
    --------
    clingo.control.Control.register_observer

    Notes
    -----
    Not all functions the `Observer` interface have to be implemented and can
    be omitted if not needed.
    '''
    def init_program(self, incremental: bool) -> None:
        '''
        Called once in the beginning.

        Parameters
        ----------
        incremental
            Whether the program is incremental. If the incremental flag is
            true, there can be multiple calls to `clingo.control.Control.solve`.
        '''

    def begin_step(self) -> None:
        '''
        Marks the beginning of a block of directives passed to the solver.
        '''

    def rule(self, choice: bool, head: Sequence[int], body: Sequence[int]) -> None:
        '''
        Observe rules passed to the solver.

        Parameters
        ----------
        choice
            Determines if the head is a choice or a disjunction.
        head
            List of program atoms forming the rule head.
        body
            List of program literals forming the rule body.
        '''

    def weight_rule(self, choice: bool, head: Sequence[int], lower_bound: int,
                    body: Sequence[Tuple[int,int]]) -> None:
        '''
        Observe rules with one weight constraint in the body passed to the
        solver.

        Parameters
        ----------
        choice
            Determines if the head is a choice or a disjunction.
        head
            List of program atoms forming the head of the rule.
        lower_bound
            The lower bound of the weight constraint in the rule body.
        body
            List of weighted literals (pairs of literal and weight) forming the
            elements of the weight constraint.
        '''

    def minimize(self, priority: int, literals: Sequence[Tuple[int,int]]) -> None:
        '''
        Observe minimize directives (or weak constraints) passed to the
        solver.

        Parameters
        ----------
        priority
            The priority of the directive.
        literals
            List of weighted literals whose sum to minimize (pairs of literal
            and weight).
        '''

    def project(self, atoms: Sequence[int]) -> None:
        '''
        Observe projection directives passed to the solver.

        Parameters
        ----------
        atoms
            The program atoms to project on.
        '''

    def output_atom(self, symbol: Symbol, atom: int) -> None:
        '''
        Observe shown atoms passed to the solver.  Facts do not have an
        associated program atom. The value of the atom is set to zero.

        Parameters
        ----------
        symbol
            The symbolic representation of the atom.
        atom
            The associated program atom (`0` for facts).
        '''

    def output_term(self, symbol: Symbol, condition: Sequence[int]) -> None:
        '''
        Observe shown terms passed to the solver.

        Parameters
        ----------
        symbol
            The symbolic representation of the term.
        condition
            List of program literals forming the condition when to show the
            term.
        '''

    def output_csp(self, symbol: Symbol, value: int,
                   condition: Sequence[int]) -> None:
        '''
        Observe shown csp variables passed to the solver.

        Parameters
        ----------
        symbol
            The symbolic representation of the variable.
        value
            The integer value of the variable.
        condition
            List of program literals forming the condition when to show the
            variable with its value.
        '''

    def external(self, atom: int, value: TruthValue) -> None:
        '''
        Observe external statements passed to the solver.

        Parameters
        ----------
        atom
            The external atom in form of a program literal.
        value
            The truth value of the external statement.
        '''

    def assume(self, literals: Sequence[int]) -> None:
        '''
        Observe assumption directives passed to the solver.

        Parameters
        ----------
        literals
            The program literals to assume (positive literals are true and
            negative literals false for the next solve call).
        '''

    def heuristic(self, atom: int, type_: HeuristicType, bias: int,
                  priority: int, condition: Sequence[int]) -> None:
        '''
        Observe heuristic directives passed to the solver.

        Parameters
        ----------
        atom
            The program atom heuristically modified.
        type_
            The type of the modification.
        bias
            A signed integer.
        priority
            An unsigned integer.
        condition
            List of program literals.
        '''

    def acyc_edge(self, node_u: int, node_v: int,
                  condition: Sequence[int]) -> None:
        '''
        Observe edge directives passed to the solver.

        Parameters
        ----------
        node_u
            The start vertex of the edge (in form of an integer).
        node_v
            Тhe end vertex of the edge (in form of an integer).
        condition
            The list of program literals forming th condition under which to
            add the edge.
        '''

    def theory_term_number(self, term_id: int, number: int) -> None:
        '''
        Observe numeric theory terms.

        Parameters
        ----------
        term_id
            The id of the term.
        number
            The value of the term.
        '''

    def theory_term_string(self, term_id : int, name : str) -> None:
        '''
        Observe string theory terms.

        Parameters
        ----------
        term_id
            The id of the term.
        name
            The string value of the term.
        '''

    def theory_term_compound(self, term_id: int, name_id_or_type: int,
                             arguments: Sequence[int]) -> None:
        '''
        Observe compound theory terms.

        Parameters
        ----------
        term_id
            The id of the term.
        name_id_or_type
            The name id or type of the term where the value `-1` stands for
            tuples, `-2` for sets, `-3` for lists, or otherwise for the id of
            the name (in form of a string term).
        arguments
            The arguments of the term in form of a list of term ids.
        '''

    def theory_element(self, element_id: int, terms: Sequence[int],
                       condition: Sequence[int]) -> None:
        '''
        Observe theory elements.

        Parameters
        ----------
        element_id
            The id of the element.
        terms
            The term tuple of the element in form of a list of term ids.
        condition
            The list of program literals forming the condition.
        '''

    def theory_atom(self, atom_id_or_zero: int, term_id: int,
                    elements: Sequence[int]) -> None:
        '''
        Observe theory atoms without guard.

        Parameters
        ----------
        atom_id_or_zero
            The id of the atom or zero for directives.
        term_id
            The term associated with the atom.
        elements
            The elements of the atom in form of a list of element ids.
        '''

    def theory_atom_with_guard(self, atom_id_or_zero: int, term_id: int,
                               elements: Sequence[int], operator_id: int,
                               right_hand_side_id: int) -> None:
        '''
        Observe theory atoms with guard.

        Parameters
        ----------
        atom_id_or_zero
            The id of the atom or zero for directives.
        term_id : int
            The term associated with the atom.
        elements
            The elements of the atom in form of a list of element ids.
        operator_id
            The id of the operator (a string term).
        right_hand_side_id
            The id of the term on the right hand side of the atom.
        '''

    def end_step(self) -> None:
        '''
        Marks the end of a block of directives passed to the solver.

        This function is called right before solving starts.
        '''

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_init_program')
def _pyclingo_observer_init_program(incremental, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.init_program(incremental)
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_begin_step')
def _pyclingo_observer_begin_step(data):
    observer: Observer = _ffi.from_handle(data).data
    observer.begin_step()
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_end_step')
def _pyclingo_observer_end_step(data):
    observer: Observer = _ffi.from_handle(data).data
    observer.end_step()
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_rule')
def _pyclingo_observer_rule(choice, head, head_size, body, body_size, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.rule(
        choice,
        [ head[i] for i in range(head_size) ],
        [ body[i] for i in range(body_size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_weight_rule')
def _pyclingo_observer_weight_rule(choice, head, head_size, lower_bound, body, body_size, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.weight_rule(
        choice,
        [ head[i] for i in range(head_size) ],
        lower_bound,
        [ (body[i].literal, body[i].weight) for i in range(body_size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_minimize')
def _pyclingo_observer_minimize(priority, literals, size, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.minimize(priority, [ (literals[i].literal, literals[i].weight) for i in range(size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_project')
def _pyclingo_observer_project(atoms, size, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.project([ atoms[i] for i in range(size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_output_atom')
def _pyclingo_observer_output_atom(symbol, atom, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.output_atom(Symbol(symbol), atom)
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_output_term')
def _pyclingo_observer_output_term(symbol, condition, size, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.output_term(Symbol(symbol), [ condition[i] for i in range(size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_output_csp')
def _pyclingo_observer_output_csp(symbol, value, condition, size, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.output_csp(Symbol(symbol), value, [ condition[i] for i in range(size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_external')
def _pyclingo_observer_external(atom, type_, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.external(atom, TruthValue(type_))
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_assume')
def _pyclingo_observer_assume(literals, size, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.assume([ literals[i] for i in range(size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_heuristic')
def _pyclingo_observer_heuristic(atom, type_, bias, priority, condition, size, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.heuristic(atom, HeuristicType(type_), bias, priority, [ condition[i] for i in range(size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_acyc_edge')
def _pyclingo_observer_acyc_edge(node_u, node_v, condition, size, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.acyc_edge(node_u, node_v, [ condition[i] for i in range(size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_theory_term_number')
def _pyclingo_observer_theory_term_number(term_id, number, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.theory_term_number(term_id, number)
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_theory_term_string')
def _pyclingo_observer_theory_term_string(term_id, name, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.theory_term_string(term_id, _to_str(name))
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_theory_term_compound')
def _pyclingo_observer_theory_term_compound(term_id, name_id_or_type, arguments, size, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.theory_term_compound(term_id, name_id_or_type, [ arguments[i] for i in range(size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_theory_element')
def _pyclingo_observer_theory_element(element_id, terms, terms_size, condition, condition_size, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.theory_element(
        element_id,
        [ terms[i] for i in range(terms_size) ],
        [ condition[i] for i in range(condition_size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_theory_atom')
def _pyclingo_observer_theory_atom(atom_id_or_zero, term_id, elements, size, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.theory_atom(atom_id_or_zero, term_id, [ elements[i] for i in range(size) ])
    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_observer_theory_atom_with_guard')
def _pyclingo_observer_theory_atom_with_guard(atom_id_or_zero, term_id, elements, size,
                                            operator_id, right_hand_side_id, data):
    observer: Observer = _ffi.from_handle(data).data
    observer.theory_atom_with_guard(
        atom_id_or_zero, term_id,
        [ elements[i] for i in range(size) ],
        operator_id, right_hand_side_id)
    return True

class Backend(ContextManager['Backend']):
    '''
    Backend object providing a low level interface to extend a logic program.

    This class allows for adding statements in ASPIF format.

    See Also
    --------
    clingo.control.Control.backend

    Notes
    -----
    The `Backend` is a context manager and must be used with Python's `with`
    statement.
    '''
    def __init__(self, rep, error):
        self._rep = rep
        self._error = error

    def __enter__(self):
        self._error.clear()
        _handle_error(_lib.clingo_backend_begin(self._rep), handler=self._error)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._error.clear()
        _handle_error(_lib.clingo_backend_end(self._rep), handler=self._error)
        return False

    def add_acyc_edge(self, node_u: int, node_v: int, condition: Sequence[int]) -> None:
        '''
        Add an edge directive to the program.

        Parameters
        ----------
        node_u
            The start node represented as an unsigned integer.
        node_v
            The end node represented as an unsigned integer.
        condition
            List of program literals.
        '''
        _handle_error(_lib.clingo_backend_acyc_edge(self._rep, node_u, node_v, condition, len(condition)))

    def add_assume(self, literals: Sequence[int]) -> None:
        '''
        Add assumptions to the program.

        Parameters
        ----------
        literals
            The list of literals to assume true.
        '''
        _handle_error(_lib.clingo_backend_assume(self._rep, literals, len(literals)))

    def add_atom(self, symbol: Optional[Symbol]=None) -> int:
        '''
        Return a fresh program atom or the atom associated with the given symbol.

        If the given symbol does not exist in the atom base, it is added first. Such
        atoms will be used in subequents calls to ground for instantiation.

        Parameters
        ----------
        symbol
            The symbol associated with the atom.
        Returns
        -------
        The program atom representing the atom.
        '''
        # pylint: disable=protected-access
        if symbol is None:
            p_sym = _ffi.NULL
        else:
            p_sym = _ffi.new('clingo_symbol_t*', symbol._rep)

        self._error.clear()
        return _c_call('clingo_atom_t', _lib.clingo_backend_add_atom, self._rep, p_sym, handler=self._error)

    def add_external(self, atom : int, value : TruthValue=TruthValue.False_) -> None:
        '''
        Mark a program atom as external optionally fixing its truth value.

        Parameters
        ----------
        atom
            The program atom to mark as external.
        value
            Optional truth value.

        Notes
        -----
        Can also be used to release an external atom using `TruthValue.Release`.
        '''
        _handle_error(_lib.clingo_backend_external(self._rep, atom, value.value))

    def add_heuristic(self, atom: int, type_: HeuristicType, bias: int, priority: int,
                      condition: Sequence[int]) -> None:
        '''
        Add a heuristic directive to the program.

        Parameters
        ----------
        atom
            Program atom to heuristically modify.
        type_
            The type of modification.
        bias
            A signed integer.
        priority
            An unsigned integer.
        condition
            List of program literals.
        '''
        _handle_error(_lib.clingo_backend_heuristic(self._rep, atom, type_.value, bias, priority,
                      condition, len(condition)))

    def add_minimize(self, priority: int, literals: Sequence[Tuple[int,int]]) -> None:
        '''
        Add a minimize constraint to the program.

        Parameters
        ----------
        priority
            Integer for the priority.
        literals
            List of pairs of program literals and weights.
        '''
        _handle_error(_lib.clingo_backend_minimize(self._rep, priority, literals, len(literals)))

    def add_project(self, atoms: Sequence[int]) -> None:
        '''
        Add a project statement to the program.

        Parameters
        ----------
        atoms
            List of program atoms to project on.
        '''
        _handle_error(_lib.clingo_backend_project(self._rep, atoms, len(atoms)))

    def add_rule(self, head: Sequence[int], body: Sequence[int]=[], choice: bool=False) -> None:
        '''
        Add a disjuntive or choice rule to the program.

        Parameters
        ----------
        head
            The program atoms forming the rule head.
        body
            The program literals forming the rule body.
        choice
            Whether to add a disjunctive or choice rule.

        Notes
        -----
        Integrity constraints and normal rules can be added by using an empty or
        singleton head list, respectively.
        '''
        # pylint: disable=dangerous-default-value
        _handle_error(_lib.clingo_backend_rule(self._rep, choice, head, len(head), body, len(body)))

    def add_weight_rule(self, head: Sequence[int], lower: int, body: Sequence[Tuple[int,int]],
                        choice: bool=False) -> None:
        '''
        Add a disjuntive or choice rule with one weight constraint with a lower bound
        in the body to the program.

        Parameters
        ----------
        head
            The program atoms forming the rule head.
        lower
            The lower bound.
        body
            The pairs of program literals and weights forming the elements of the
            weight constraint.
        choice
            Whether to add a disjunctive or choice rule.
        '''
        _handle_error(_lib.clingo_backend_weight_rule(self._rep, choice, head, len(head), lower, body, len(body)))



'''
Functions and classes related to configuration.

Examples
--------
The following example shows how to modify the configuration to enumerate all
models:

```python
>>> from clingo.control import Control
>>>
>>> ctl = Control()
>>> ctl.configuration.keys
['tester', 'solve', 'asp', 'solver', 'configuration', 'share',
 'learn_explicit', 'sat_prepro', 'stats', 'parse_ext', 'parse_maxsat']
>>> ctl.configuration.solve.keys
['solve_limit', 'parallel_mode', 'global_restarts', 'distribute',
 'integrate', 'enum_mode', 'project', 'models', 'opt_mode']
>>> ctl.configuration.solve.description("models")
'Compute at most %A models (0 for all)\\n'
>>> ctl.configuration.solve.models = 0
>>> ctl.add("base", [], "1 {a; b}.")
>>> ctl.ground([("base", [])])
>>> print(ctl.solve(on_model=print))
b
a
a b
SAT
```
'''

from typing import List, Optional, Union

from ._internal import _c_call, _ffi, _handle_error, _lib, _to_str

__all__ = [ 'Configuration' ]

class Configuration:
    '''
    Allows for changing the configuration of the underlying solver.

    Options are organized hierarchically. To change and inspect an option use:

        config.group.subgroup.option = "value"
        value = config.group.subgroup.option

    There are also arrays of option groups that can be accessed using integer
    indices:

        config.group.subgroup[0].option = "value1"
        config.group.subgroup[1].option = "value2"

    To list the subgroups of an option group, use the `Configuration.keys`
    member. Array option groups, like solver, can be iterated. Furthermore,
    there are meta options having key `configuration`. Assigning a meta option
    sets a number of related options.  To get further information about an
    option or option group, use `Configuration.description`.

    Notes
    -----
    The value of an option is always a string and any value assigned to an
    option is automatically converted into a string.

    See Also
    --------
    clingo.control.Control.configuration
    '''
    def __init__(self, rep, key):
        # Note: we have to bypass __setattr__ to avoid infinite recursion
        super().__setattr__("_rep", rep)
        super().__setattr__("_key", key)

    @property
    def _type(self) -> int:
        return _c_call('clingo_configuration_type_bitset_t', _lib.clingo_configuration_type, self._rep, self._key)

    @property
    def is_array(self) -> bool:
        '''
        This property is true if the configuration option is an array.
        '''
        return bool(self._type & _lib.clingo_configuration_type_array)

    def _get_subkey(self, name: str) -> Optional[int]:
        if self._type & _lib.clingo_configuration_type_map:
            if _c_call('bool', _lib.clingo_configuration_map_has_subkey, self._rep, self._key, name.encode()):
                return _c_call('clingo_id_t', _lib.clingo_configuration_map_at, self._rep, self._key, name.encode())
        return None

    def __len__(self):
        if self._type & _lib.clingo_configuration_type_array:
            return _c_call('size_t', _lib.clingo_configuration_array_size, self._rep, self._key)
        return 0

    def __getitem__(self, idx: int) -> 'Configuration':
        if idx < 0 or idx >= len(self):
            raise IndexError("invalid index")

        key = _c_call('clingo_id_t', _lib.clingo_configuration_array_at, self._rep, self._key, idx)
        return Configuration(self._rep, key)

    def __getattr__(self, name: str) -> Union[None, str, 'Configuration']:
        key = self._get_subkey(name)
        if key is None:
            raise AttributeError(f'no attribute: {name}')

        type_ = _c_call('clingo_configuration_type_bitset_t', _lib.clingo_configuration_type, self._rep, key)

        if type_ & _lib.clingo_configuration_type_value:
            if not _c_call('bool', _lib.clingo_configuration_value_is_assigned, self._rep, key):
                return None

            size = _c_call('size_t', _lib.clingo_configuration_value_get_size, self._rep, key)

            c_val = _ffi.new('char[]', size)
            _handle_error(_lib.clingo_configuration_value_get(self._rep, key, c_val, size))
            return _to_str(c_val)

        return Configuration(self._rep, key)

    def __setattr__(self, name: str, val) -> None:
        key = self._get_subkey(name)
        if key is None:
            super().__setattr__(name, val)
        else:
            _handle_error(_lib.clingo_configuration_value_set(self._rep, key, str(val).encode()))

    def description(self, name: str) -> str:
        '''
        Get a description for an option or option group.

        Parameters
        ----------
        name
            The name of the option.
        '''
        key = self._get_subkey(name)
        if key is None:
            raise RuntimeError(f'unknown option {name}')
        return _to_str(_c_call('char*', _lib.clingo_configuration_description, self._rep, key))

    @property
    def keys(self) -> Optional[List[str]]:
        '''
        The list of names of sub-option groups or options.

        The list is `None` if the current object is not an option group.
        '''
        ret = None
        if self._type & _lib.clingo_configuration_type_map:
            ret = []
            for i in range(_c_call('size_t', _lib.clingo_configuration_map_size, self._rep, self._key)):
                name = _c_call('char*', _lib.clingo_configuration_map_subkey_name, self._rep, self._key, i)
                ret.append(_to_str(name))
        return ret



"""
Utility functions and classes.
"""

from typing import Optional, cast
from collections import abc

# pylint: disable=too-many-ancestors

class Slice:
    '''
    Wrapper for Python's slice that computes index ranges to slice sequences.

    Currently, the range is recomputed each time. It is probably also possible
    to combine the involved slices into one.
    '''
    def __init__(self, slc: slice, rec: Optional['Slice']=None):
        self._slc = slc
        self._rec = rec

    def rng(self, size):
        '''
        Return a range providing indices to access a sequence of length size.
        '''
        return (range(*self._slc.indices(size))
                if self._rec is None else
                self._rec.rng(size)[self._slc])

class SlicedSequence(abc.Sequence):
    '''
    Helper to slice sequences.
    '''
    def __init__(self, seq: abc.Sequence, slc: Slice):
        self._seq = seq
        self._slc = slc
        self._len = -1
        self._lst = None

    @property
    def _rng(self):
        size = len(self._seq)
        if size != self._len:
            self._lst = self._slc.rng(size)
            self._len = size
        return self._lst

    def __len__(self) -> int:
        return len(self._rng)

    def __iter__(self):
        for idx in self._rng:
            yield self._seq[idx]

    def __getitem__(self, slc):
        if isinstance(slc, slice):
            return SlicedSequence(self._seq, Slice(slc, self._slc))
        return self._seq[self._rng[slc]]

    def __str__(self):
        return str(list(self))

    def __repr__(self):
        return repr(list(self))


class SlicedMutableSequence(SlicedSequence, abc.MutableSequence):
    '''
    Helper to slice sequences.
    '''
    def __init__(self, seq: abc.MutableSequence, slc: Slice):
        super().__init__(seq, slc)

    @property
    def _mut_seq(self):
        return cast(abc.MutableSequence, self._seq)

    def __setitem__(self, index, ast):
        if isinstance(index, slice):
            raise TypeError('slicing not implemented')
        self._mut_seq[self._rng[index]] = ast

    def __delitem__(self, index):
        if isinstance(index, slice):
            raise TypeError('slicing not implemented')
        del self._mut_seq[self._rng[index]]

    def insert(self, index, value):
        self._mut_seq[self._rng[index]] = value



'''
This module contains the `clingo.control.Control` class responsible for
controling grounding and solving.

Examples
--------
The following example shows basic (multishot) grounding and solving.

    >>> from clingo.symbol import Number
    >>> from clingo.control import Control
    >>>
    >>> ctl = Control()
    >>> ctl.add("base", [], "q.")
    >>> ctl.add("p", ["t"], "q(t).")
    >>>
    >>> ctl.ground([("base", [])])
    >>> print(ctl.solve(on_model=print))
    q
    SAT
    >>> ctl.ground([("p", [Number(1)]), ("p", [Number(2)])])
    >>> print(ctl.solve(on_model=print))
    q q(1) q(2)
    SAT
    >>> ctl.ground([("p", [Number(3)])])
    >>> print(ctl.solve(on_model=print))
    q q(1) q(2) q(3)
    SAT
'''

from typing import Any, Callable, Iterator, Optional, Sequence, Tuple, Union, cast
from collections import abc
import sys

from ._internal import _CBData, _Error, _cb_error_handler, _c_call, _ffi, _handle_error, _lib, _overwritten
from .core import Logger
from .symbol import Symbol
from .symbolic_atoms import SymbolicAtoms
from .theory_atoms import TheoryAtom
from .solving import Model, SolveHandle, SolveResult
from .propagator import Propagator
from .backend import Backend, Observer
from .configuration import Configuration
from .statistics import StatisticsMap, _mutable_statistics, _statistics

__all__ = [ 'Control' ]

class _SolveEventHandler:
    # pylint: disable=missing-function-docstring
    def __init__(self, on_model, on_unsat, on_statistics, on_finish):
        self._on_model = on_model
        self._on_unsat = on_unsat
        self._on_statistics = on_statistics
        self._on_finish = on_finish

    def on_model(self, m):
        ret = None
        if self._on_model is not None:
            ret = self._on_model(Model(m))
        return bool(ret or ret is None)

    def on_unsat(self, lower):
        if self._on_unsat is not None:
            self._on_unsat(lower)

    def on_finish(self, res):
        if self._on_finish is not None:
            self._on_finish(SolveResult(res))

    def on_statistics(self, step, accu):
        if self._on_statistics is not None:
            self._on_statistics(_mutable_statistics(step), _mutable_statistics(accu))

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_solve_event_callback')
def _pyclingo_solve_event_callback(type_, event, data, goon):
    '''
    Low-level solve event handler.
    '''
    handler = _ffi.from_handle(data).data
    if type_ == _lib.clingo_solve_event_type_finish:
        handler.on_finish(_ffi.cast('clingo_solve_result_bitset_t*', event)[0])

    if type_ == _lib.clingo_solve_event_type_model:
        goon[0] = handler.on_model(_ffi.cast('clingo_model_t*', event))

    if type_ == _lib.clingo_solve_event_type_unsat:
        c_args = _ffi.cast('void**', event)
        c_lower = _ffi.cast('int64_t*', c_args[0])
        size = int(_ffi.cast('size_t', c_args[1]))
        handler.on_unsat([c_lower[i] for i in range(size)])

    if type_ == _lib.clingo_solve_event_type_statistics:
        p_stats = _ffi.cast('clingo_statistics_t**', event)
        handler.on_statistics(p_stats[0], p_stats[1])

    return True

@_ffi.def_extern(onerror=_cb_error_handler('data'), name='pyclingo_ground_callback')
def _pyclingo_ground_callback(location, name, arguments, arguments_size, data, symbol_callback, symbol_callback_data):
    '''
    Low-level ground callback.
    '''
    # Note: location could be attached to error message
    # pylint: disable=protected-access,unused-argument
    context = _ffi.from_handle(data).data
    py_name = _ffi.string(name).decode()
    fun = getattr(sys.modules['__main__'] if context is None else context, py_name)

    args = []
    for i in range(arguments_size):
        args.append(Symbol(arguments[i]))

    ret = fun(*args)
    symbols = list(ret) if isinstance(ret, abc.Iterable) else [ret]

    c_symbols = _ffi.new('clingo_symbol_t[]', len(symbols))
    for i, sym in enumerate(symbols):
        c_symbols[i] = sym._rep
    _handle_error(symbol_callback(c_symbols, len(symbols), symbol_callback_data))

    return True

class Control:
    '''
    Control object for the grounding/solving process.

    Parameters
    ----------
    arguments
        Arguments to the grounder and solver.
    logger
        Function to intercept messages normally printed to standard error.
    message_limit
        The maximum number of messages passed to the logger.

    Notes
    -----
    Note that only gringo options (without `--text`) and clasp's search options
    are supported. Furthermore, you must not call any functions of a `Control`
    object while a solve call is active.
    '''
    def __init__(self, arguments: Sequence[str]=[],
                 logger: Optional[Logger]=None, message_limit: int=20):
        # pylint: disable=protected-access,dangerous-default-value
        self._free = False
        self._mem = []
        if isinstance(arguments, abc.Sequence):
            if logger is not None:
                c_handle = _ffi.new_handle(logger)
                c_cb = _lib.pyclingo_logger_callback
                self._mem.append(c_handle)
            else:
                c_handle = _ffi.NULL
                c_cb = _ffi.NULL
            c_mem = []
            c_args = _ffi.new('char*[]', len(arguments))
            for i, arg in enumerate(arguments):
                c_mem.append(_ffi.new("char[]", arg.encode()))
                c_args[i] = c_mem[-1]
            self._rep = _c_call('clingo_control_t *', _lib.clingo_control_new,
                                c_args, len(arguments), c_cb, c_handle, message_limit)
            self._free = True
        else:
            self._rep = arguments

        self._handler = None
        self._statistics = None
        self._statistics_call = -1.0
        self._error = _Error()

    def __del__(self):
        if self._free:
            _lib.clingo_control_free(self._rep)

    def add(self, name: str, parameters: Sequence[str], program: str) -> None:
        '''
        Extend the logic program with the given non-ground logic program in string form.

        Parameters
        ----------
        name
            The name of program block to add.
        parameters
            The parameters of the program block to add.
        program
            The non-ground program in string form.

        See Also
        --------
        Control.ground
        '''
        c_mem = []
        c_params = _ffi.new('char*[]', len(parameters))
        for i, param in enumerate(parameters):
            c_mem.append(_ffi.new("char[]", param.encode()))
            c_params[i] = c_mem[-1]
        _handle_error(_lib.clingo_control_add(self._rep, name.encode(), c_params, len(parameters), program.encode()))

    def _program_atom(self, lit: Union[Symbol,int]) -> int:
        if isinstance(lit, int):
            return lit
        satom = self.symbolic_atoms[lit]
        return 0 if satom is None else satom.literal

    def assign_external(self, external: Union[Symbol,int], truth: Optional[bool]) -> None:
        '''
        Assign a truth value to an external atom.

        Parameters
        ----------
        external
            A symbol or program literal representing the external atom.
        truth
            A Boolean fixes the external to the respective truth value; and
            None leaves its truth value open.

        See Also
        --------
        Control.release_external, clingo.solving.SolveControl.symbolic_atoms,
        clingo.symbolic_atoms.SymbolicAtom.is_external

        Notes
        -----
        The truth value of an external atom can be changed before each solve
        call. An atom is treated as external if it has been declared using an
        `#external` directive, and has not been released by calling
        `Control.release_external` or defined in a logic program with some
        rule. If the given atom is not external, then the function has no
        effect.

        For convenience, the truth assigned to atoms over negative program
        literals is inverted.
        '''

        if truth is None:
            val = _lib.clingo_external_type_free
        elif truth:
            val = _lib.clingo_external_type_true
        else:
            val = _lib.clingo_external_type_false
        _handle_error(_lib.clingo_control_assign_external(self._rep, self._program_atom(external), val))

    def backend(self) -> Backend:
        '''
        Returns a `Backend` object providing a low level interface to extend a
        logic program.

        See Also
        --------
        clingo.backend
        '''
        return Backend(_c_call('clingo_backend_t*', _lib.clingo_control_backend, self._rep), self._error)

    def cleanup(self) -> None:
        '''
        Cleanup the domain used for grounding by incorporating information from
        the solver.

        This function cleans up the domain used for grounding.  This is done by
        first simplifying the current program representation (falsifying
        released external atoms).  Afterwards, the top-level implications are
        used to either remove atoms from the domain or mark them as facts.

        See Also
        --------
        Control.enable_cleanup

        Notes
        -----
        Any atoms falsified are completely removed from the logic program.
        Hence, a definition for such an atom in a successive step introduces a
        fresh atom.

        With the current implementation, the function only has an effect if
        called after solving and before any function is called that starts a
        new step.

        Typically, it is not necessary to call this function manually because
        automatic cleanups are enabled by default.
        '''
        _handle_error(_lib.clingo_control_cleanup(self._rep))

    def get_const(self, name: str) -> Optional[Symbol]:
        '''
        Return the symbol for a constant definition of form:

            #const name = symbol.

        Parameters
        ----------
        name
            The name of the constant to retrieve.

        Returns
        -------
        The function returns `None` if no matching constant definition exists.
        '''
        if not _c_call('bool', _lib.clingo_control_has_const, self._rep, name.encode()):
            return None

        return Symbol(_c_call('clingo_symbol_t', _lib.clingo_control_get_const, self._rep, name.encode()))

    def ground(self, parts: Sequence[Tuple[str,Sequence[Symbol]]], context: Any=None) -> None:
        '''
        Ground the given list of program parts specified by tuples of names and
        arguments.

        Parameters
        ----------
        parts
            List of tuples of program names and program arguments to ground.
        context
            A context object whose methods are called during grounding using
            the `@`-syntax (if omitted, those from the main module are used).

        Notes
        -----
        Note that parts of a logic program without an explicit `#program`
        specification are by default put into a program called `base` without
        arguments.
        '''
        # pylint: disable=protected-access,dangerous-default-value
        self._error.clear()
        data = _CBData(context, self._error)
        c_data = _ffi.new_handle(data) if context else _ffi.NULL
        c_cb = _lib.pyclingo_ground_callback if context else _ffi.NULL

        c_mem = []
        c_parts = _ffi.new("clingo_part_t[]", len(parts))
        for part, c_part in zip(parts, c_parts):
            c_mem.append(_ffi.new("char[]", part[0].encode()))
            c_part.name = c_mem[-1]
            c_mem.append(_ffi.new("clingo_symbol_t[]", len(part[1])))
            c_part.params = c_mem[-1]
            for i, sym in enumerate(part[1]):
                c_part.params[i] = sym._rep
            c_part.size = len(part[1])

        _handle_error(_lib.clingo_control_ground(
            self._rep, c_parts, len(parts), c_cb, c_data), data)

    def interrupt(self) -> None:
        '''
        Interrupt the active solve call.

        Notes
        -----
        This function is thread-safe and can be called from a signal handler. If no
        search is active, the subsequent call to `Control.solve` is interrupted. The
        result of the `Control.solve` method can be used to query if the search was
        interrupted.
        '''
        _lib.clingo_control_interrupt(self._rep)

    def load(self, path: str) -> None:
        '''
        Extend the logic program with a (non-ground) logic program in a file.

        Parameters
        ----------
        path
            The path of the file to load.
        '''
        _handle_error(_lib.clingo_control_load(self._rep, path.encode()))

    def register_observer(self, observer: Observer, replace: bool=False) -> None:
        '''
        Registers the given observer to inspect the produced grounding.

        Parameters
        ----------
        observer
            The observer to register. See below for a description of the requirede
            interface.
        replace
            If set to true, the output is just passed to the observer and nolonger to
            the underlying solver (or any previously registered observers).

        See Also
        --------
        clingo.backend
        '''
        # pylint: disable=protected-access,line-too-long
        c_observer = _ffi.new('clingo_ground_program_observer_t*', (
            _lib.pyclingo_observer_init_program if _overwritten(Observer, observer, "init_program") else _ffi.NULL,
            _lib.pyclingo_observer_begin_step if _overwritten(Observer, observer, "begin_step") else _ffi.NULL,
            _lib.pyclingo_observer_end_step if _overwritten(Observer, observer, "end_step") else _ffi.NULL,
            _lib.pyclingo_observer_rule if _overwritten(Observer, observer, "rule") else _ffi.NULL,
            _lib.pyclingo_observer_weight_rule if _overwritten(Observer, observer, "weight_rule") else _ffi.NULL,
            _lib.pyclingo_observer_minimize if _overwritten(Observer, observer, "minimize") else _ffi.NULL,
            _lib.pyclingo_observer_project if _overwritten(Observer, observer, "project") else _ffi.NULL,
            _lib.pyclingo_observer_output_atom if _overwritten(Observer, observer, "output_atom") else _ffi.NULL,
            _lib.pyclingo_observer_output_term if _overwritten(Observer, observer, "output_term") else _ffi.NULL,
            _lib.pyclingo_observer_output_csp if _overwritten(Observer, observer, "output_csp") else _ffi.NULL,
            _lib.pyclingo_observer_external if _overwritten(Observer, observer, "external") else _ffi.NULL,
            _lib.pyclingo_observer_assume if _overwritten(Observer, observer, "assume") else _ffi.NULL,
            _lib.pyclingo_observer_heuristic if _overwritten(Observer, observer, "heuristic") else _ffi.NULL,
            _lib.pyclingo_observer_acyc_edge if _overwritten(Observer, observer, "acyc_edge") else _ffi.NULL,
            _lib.pyclingo_observer_theory_term_number if _overwritten(Observer, observer, "theory_term_number") else _ffi.NULL,
            _lib.pyclingo_observer_theory_term_string if _overwritten(Observer, observer, "theory_term_string") else _ffi.NULL,
            _lib.pyclingo_observer_theory_term_compound if _overwritten(Observer, observer, "theory_term_compound") else _ffi.NULL,
            _lib.pyclingo_observer_theory_element if _overwritten(Observer, observer, "theory_element") else _ffi.NULL,
            _lib.pyclingo_observer_theory_atom if _overwritten(Observer, observer, "theory_atom") else _ffi.NULL,
            _lib.pyclingo_observer_theory_atom_with_guard if _overwritten(Observer, observer, "theory_atom_with_guard") else _ffi.NULL))
        c_data = _ffi.new_handle(_CBData(observer, self._error))
        self._mem.append(c_data)
        _handle_error(_lib.clingo_control_register_observer(self._rep, c_observer, replace, c_data))

    def register_propagator(self, propagator: Propagator) -> None:
        '''
        Registers the given propagator with all solvers.

        Parameters
        ----------
        propagator
            The propagator to register.

        See Also
        --------
        clingo.propagator
        '''
        # pylint: disable=protected-access
        c_propagator = _ffi.new('clingo_propagator_t*', (
            _lib.pyclingo_propagator_init if _overwritten(Propagator, propagator, "init") else _ffi.NULL,
            _lib.pyclingo_propagator_propagate if _overwritten(Propagator, propagator, "propagate") else _ffi.NULL,
            _lib.pyclingo_propagator_undo if _overwritten(Propagator, propagator, "undo") else _ffi.NULL,
            _lib.pyclingo_propagator_check if _overwritten(Propagator, propagator, "check") else _ffi.NULL,
            _lib.pyclingo_propagator_decide if _overwritten(Propagator, propagator, "decide") else _ffi.NULL))
        c_data = _ffi.new_handle(_CBData(propagator, self._error))
        self._mem.append(c_data)
        _handle_error(_lib.clingo_control_register_propagator(self._rep, c_propagator, c_data, False))

    def release_external(self, external: Union[Symbol,int]) -> None:
        '''
        Release an external atom represented by the given symbol or program
        literal.

        This function causes the corresponding atom to become permanently false
        if there is no definition for the atom in the program. Otherwise, the
        function has no effect.

        Parameters
        ----------
        external
            The symbolic atom or program atom to release.

        Notes
        -----
        If the program literal is negative, the corresponding atom is released.

        Examples
        --------
        The following example shows the effect of assigning and releasing and external
        atom.

            >>> from clingo.symbol import Function
            >>> from clingo.control import Control
            >>>
            >>> ctl = Control()
            >>> ctl.add("base", [], "a. #external b.")
            >>> ctl.ground([("base", [])])
            >>> ctl.assign_external(Function("b"), True)
            >>> print(ctl.solve(on_model=print))
            b a
            SAT
            >>> ctl.release_external(Function("b"))
            >>> print(ctl.solve(on_model=print))
            a
            SAT
        '''
        _handle_error(_lib.clingo_control_release_external(self._rep, self._program_atom(external)))

    def solve(self,
              assumptions: Sequence[Union[Tuple[Symbol,bool],int]]=[],
              on_model: Optional[Callable[[Model],Optional[bool]]]=None,
              on_unsat: Optional[Callable[[Sequence[int]],None]]=None,
              on_statistics : Optional[Callable[[StatisticsMap,StatisticsMap],None]]=None,
              on_finish: Optional[Callable[[SolveResult],None]]=None,
              on_core: Optional[Callable[[Sequence[int]],None]]=None,
              yield_: bool=False,
              async_: bool=False) -> Union[SolveHandle,SolveResult]:
        '''
        Starts a search.

        Parameters
        ----------
        assumptions
            List of (atom, boolean) tuples or program literals (see
            `clingo.symbolic_atoms.SymbolicAtom.literal`) that serve as
            assumptions for the solve call, e.g., solving under assumptions
            `[(Function("a"), True)]` only admits answer sets that contain atom `a`.
        on_model
            Optional callback for intercepting models.
            A `clingo.solving.Model` object is passed to the callback. The
            search can be interruped from the model callback by returning
            False.
        on_unsat
            Optional callback to intercept lower bounds during optimization.
        on_statistics
            Optional callback to update statistics.
            The step and accumulated statistics are passed as arguments.
        on_finish
            Optional callback called once search has finished.
            A `clingo.solving.SolveResult` also indicating whether the solve
            call has been intrrupted is passed to the callback.
        on_core
            Optional callback called with the assumptions that made a problem
            unsatisfiable.
        yield_
            The resulting `clingo.solving.SolveHandle` is iterable yielding
            `clingo.solving.Model` objects.
        async_
            The solve call and the method `clingo.solving.SolveHandle.resume`
            of the returned handle are non-blocking.

        Returns
        -------
        The return value depends on the parameters. If either `yield_` or
        `async_` is true, then a handle is returned. Otherwise, a
        `clingo.solving.SolveResult` is returned.

        See Also
        --------
        clingo.solving

        Notes
        -----
        If neither `yield_` nor `async_` is set, the function returns a
        `clingo.solving.SolveResult` right away.

        In gringo or in clingo with lparse or text output enabled, this
        function just grounds and returns a `clingo.solving.SolveResult` where
        `clingo.solving.SolveResult.unknown` is true.

        If this function is used in embedded Python code, you might want to start
        clingo using the `--outf=3` option to disable all output from clingo.

        Asynchronous solving is only available in clingo with thread support
        enabled. Furthermore, the on_model and on_finish callbacks are called
        from another thread. To ensure that the methods can be called, make
        sure to not use any functions that block Python's GIL indefinitely.

        This function as well as blocking functions on the
        `clingo.solving.SolveHandle` release the GIL but are not thread-safe.
        '''
        # pylint: disable=protected-access,dangerous-default-value
        self._error.clear()
        handler = _SolveEventHandler(on_model, on_unsat, on_statistics, on_finish)
        data = _CBData(handler, self._error)
        self._handler = _ffi.new_handle(data)

        p_ass = _ffi.NULL
        if assumptions:
            atoms = None
            p_ass = _ffi.new('clingo_literal_t[]', len(assumptions))
            for i, lit in enumerate(assumptions):
                if isinstance(lit, int):
                    p_ass[i] = lit
                else:
                    if atoms is None:
                        atoms = self.symbolic_atoms
                    atom = self.symbolic_atoms[lit[0]]
                    slit = -1 if atom is None else atom.literal
                    p_ass[i] = slit if lit[1] else -slit

        mode = 0
        if yield_:
            mode |= _lib.clingo_solve_mode_yield
        if async_:
            mode |= _lib.clingo_solve_mode_async

        handle = SolveHandle(
            _c_call('clingo_solve_handle_t*', _lib.clingo_control_solve,
                self._rep, mode,
                p_ass, len(assumptions),
                _lib.pyclingo_solve_event_callback, self._handler,
                handler=data),
            data)

        if not yield_ and not async_:
            with handle:
                ret = handle.get()
                if on_core is not None and ret.unsatisfiable:
                    on_core(handle.core())
                return ret
        return handle

    @property
    def configuration(self) -> Configuration:
        '''
        Object to change the configuration.
        '''
        conf = _c_call('clingo_configuration_t*', _lib.clingo_control_configuration, self._rep)
        key = _c_call('clingo_id_t', _lib.clingo_configuration_root, conf)
        return Configuration(conf, key)

    @property
    def enable_cleanup(self) -> bool:
        '''
        Whether to enable automatic calls to `Control.cleanup`.
        '''
        return _lib.clingo_control_get_enable_cleanup(self._rep)

    @enable_cleanup.setter
    def enable_cleanup(self, value: bool) -> None:
        _handle_error(_lib.clingo_control_set_enable_cleanup(self._rep, value))

    @property
    def enable_enumeration_assumption(self) -> bool:
        '''
        Whether do discard or keep learnt information from enumeration modes.

        If the enumeration assumption is enabled, then all information learnt from
        clasp's various enumeration modes is removed after a solve call. This includes
        enumeration of cautious or brave consequences, enumeration of answer sets with
        or without projection, or finding optimal models; as well as clauses added with
        `clingo.solving.SolveControl.add_clause`.

        Notes
        -----
        Initially the enumeration assumption is enabled.

        In general, the enumeration assumption should be enabled whenever there are
        multiple calls to solve. Otherwise, the behavior of the solver will be
        unpredictable because there are no guarantees which information exactly is
        kept. There might be small speed benefits when disabling the enumeration
        assumption for single shot solving.
        '''
        return _lib.clingo_control_get_enable_enumeration_assumption(self._rep)

    @enable_enumeration_assumption.setter
    def enable_enumeration_assumption(self, value: bool) -> None:
        _handle_error(_lib.clingo_control_set_enable_enumeration_assumption(self._rep, value))

    @property
    def is_conflicting(self) -> bool:
        '''
        Whether the internal program representation is conflicting.

        If this (read-only) property is true, solve calls return immediately with an
        unsatisfiable solve result.

        Notes
        -----
        Conflicts first have to be detected, e.g., initial unit propagation results in
        an empty clause, or later if an empty clause is resolved during solving. Hence,
        the property might be false even if the problem is unsatisfiable.
        '''
        return _lib.clingo_control_is_conflicting(self._rep)

    @property
    def statistics(self) -> dict:
        '''
        A `dict` containing solve statistics of the last solve call.

        See Also
        --------
        clingo.statistics

        Notes
        -----
        The statistics correspond to the `--stats` output of clingo. The detail of the
        statistics depends on what level is requested on the command line. Furthermore,
        there are some functions like `Control.release_external` that start a new
        solving step resetting the current step statistics. It is best to access the
        statistics right after solving.

        This property is only available in clingo.
        '''
        stats = _c_call('clingo_statistics_t*', _lib.clingo_control_statistics, self._rep)

        p_key = _ffi.new('uint64_t*')
        key_root = _c_call(p_key, _lib.clingo_statistics_root, stats)

        key_summary = _c_call(p_key, _lib.clingo_statistics_map_at, stats, key_root, "summary".encode())
        key_call = _c_call(p_key, _lib.clingo_statistics_map_at, stats, key_summary, "call".encode())
        call = _c_call('double', _lib.clingo_statistics_value_get, stats, key_call)
        if self._statistics is not None and call != self._statistics_call:
            self._statistics = None

        if self._statistics is None:
            self._statistics_call = call
            self._statistics = _statistics(stats, key_root)

        return cast(dict, self._statistics)

    @property
    def symbolic_atoms(self) -> SymbolicAtoms:
        '''
        An object to inspect the symbolic atoms.

        See Also
        --------
        clingo.symbolic_atoms
        '''
        return SymbolicAtoms(_c_call('clingo_symbolic_atoms_t*', _lib.clingo_control_symbolic_atoms, self._rep))

    @property
    def theory_atoms(self) -> Iterator[TheoryAtom]:
        '''
        An iterator over the theory atoms in a program.

        See Also
        --------
        clingo.theory_atoms
        '''
        atoms = _c_call('clingo_theory_atoms_t*', _lib.clingo_control_theory_atoms, self._rep)
        size = _c_call('size_t', _lib.clingo_theory_atoms_size, atoms)

        for idx in range(size):
            yield TheoryAtom(atoms, idx)



'''
Tests for theory and symbolic atoms.
'''

from unittest import TestCase
from typing import cast
from clingo import Control, Function, Number, TheoryTermType
from clingo.symbolic_atoms import SymbolicAtom

THEORY = '''
#theory test {
    t { };
    &a/0 : t, head;
    &b/0 : t, {=}, t, head
}.
'''

class TestAtoms(TestCase):
    '''
    Tests for theory and symbolic atoms.
    '''

    def setUp(self):
        self.ctl = Control()

    def test_symbolic_atom(self):
        '''
        Test symbolic atom.
        '''
        self.ctl.add('base', [], 'p(1). {p(2)}. #external p(3).')
        self.ctl.ground([('base', [])])

        atoms = self.ctl.symbolic_atoms

        p1 = cast(SymbolicAtom, atoms[Function('p', [Number(1)])])
        self.assertIsNotNone(p1)
        self.assertTrue(p1.is_fact)
        self.assertFalse(p1.is_external)
        self.assertTrue(p1.literal >= 1)
        self.assertEqual(p1.symbol, Function('p', [Number(1)]))
        self.assertTrue(p1.match('p', 1, True))
        self.assertFalse(p1.match('p', 2, True))

        p2 = cast(SymbolicAtom, atoms[Function('p', [Number(2)])])
        self.assertIsNotNone(p2)
        self.assertFalse(p2.is_fact)
        self.assertFalse(p2.is_external)
        self.assertTrue(p2.literal >= 2)
        self.assertEqual(p2.symbol, Function('p', [Number(2)]))
        self.assertTrue(p2.match('p', 1, True))
        self.assertFalse(p2.match('p', 2, True))

        p3 = cast(SymbolicAtom, atoms[Function('p', [Number(3)])])
        self.assertIsNotNone(p3)
        self.assertFalse(p3.is_fact)
        self.assertTrue(p3.is_external)
        self.assertTrue(p3.literal >= 2)
        self.assertEqual(p3.symbol, Function('p', [Number(3)]))
        self.assertTrue(p3.match('p', 1, True))
        self.assertFalse(p3.match('p', 2, True))

        p4 = atoms[Function('p', [Number(4)])]
        self.assertIsNone(p4)

    def test_symbolic_atoms(self):
        '''
        Test symbolic atoms.
        '''
        self.ctl.add('base', [], 'p(1). {p(2)}. #external p(3). q(1). -p(1). {q(2)}. #external q(3).')
        self.ctl.ground([('base', [])])

        atoms = self.ctl.symbolic_atoms
        self.assertEqual(sorted(atoms.signatures), [('p', 1, False), ('p', 1, True), ('q', 1, True)])

        ps = list(atoms.by_signature('p', 1, True))
        self.assertEqual(len(ps), 3)
        for p in ps:
            self.assertEqual(p.symbol.name, 'p')
            self.assertTrue(p.symbol.positive)
            self.assertEqual(len(p.symbol.arguments), 1)

        nps = list(atoms.by_signature('p', 1, False))
        self.assertEqual(len(nps), 1)
        for p in nps:
            self.assertEqual(p.symbol.name, 'p')
            self.assertTrue(p.symbol.negative)
            self.assertEqual(p.symbol.arguments, [Number(1)])

        self.assertEqual(len(atoms), 7)
        self.assertEqual(len(list(atoms)), 7)

        self.assertIn(Function('p', [Number(1)], True), atoms)
        self.assertIn(Function('p', [Number(1)], False), atoms)
        self.assertIn(Function('q', [Number(2)], True), atoms)
        self.assertNotIn(Function('q', [Number(2)], False), atoms)

    def test_theory_term(self):
        '''
        Test theory term.
        '''
        self.ctl.add('base', [], THEORY)
        self.ctl.add('base', [], '&a { 1,a,f(a),{1},(1,),[1] }.')
        self.ctl.ground([('base', [])])

        terms = next(self.ctl.theory_atoms).elements[0].terms
        self.assertEqual([str(term) for term in terms], ['1', 'a', 'f(a)', '{1}', '(1,)', '[1]'])
        num, sym, fun, set_, tup, lst = terms
        self.assertEqual(num.type, TheoryTermType.Number)
        self.assertEqual(num.number, 1)
        self.assertEqual(sym.type, TheoryTermType.Symbol)
        self.assertEqual(sym.name, 'a')
        self.assertEqual(fun.type, TheoryTermType.Function)
        self.assertEqual(fun.name, 'f')
        self.assertEqual(fun.arguments, [sym])
        self.assertEqual(set_.type, TheoryTermType.Set)
        self.assertEqual(set_.arguments, [num])
        self.assertEqual(tup.type, TheoryTermType.Tuple)
        self.assertEqual(tup.arguments, [num])
        self.assertEqual(lst.type, TheoryTermType.List)
        self.assertEqual(lst.arguments, [num])

        self.assertNotEqual(hash(num), hash(sym))
        self.assertEqual(hash(num), hash(lst.arguments[0]))
        self.assertNotEqual(num < sym, sym < num)

        self.assertRegex(repr(num), 'TheoryTerm(.*)')

    def test_theory_element(self):
        '''
        Test theory element.
        '''
        self.ctl.add('base', [], THEORY)
        self.ctl.add('base', [], '{a; b}.')
        self.ctl.add('base', [], '&a { 1; 2,3: a,b }.')
        self.ctl.ground([('base', [])])

        atom = next(self.ctl.theory_atoms)
        elements = sorted(atom.elements, key=lambda elem: len(elem.terms))
        self.assertEqual([str(elem) for elem in elements], ['1', '2,3: a,b'])

        a, b = elements
        self.assertEqual(len(a.condition), 0)
        self.assertEqual(len(b.condition), 2)
        self.assertTrue(all(lit >= 1 for lit in b.condition))
        self.assertGreaterEqual(b.condition_id, 1)

        self.assertEqual(a, a)
        self.assertNotEqual(a, b)
        self.assertNotEqual(hash(a), hash(b))
        self.assertNotEqual(a < b, b < a)

        self.assertRegex(repr(a), 'TheoryElement(.*)')

    def test_theory_atom(self):
        '''
        Test theory atom.
        '''
        self.ctl.add('base', [], THEORY)
        self.ctl.add('base', [], '&a {}.')
        self.ctl.add('base', [], '&b {} = 1.')
        self.ctl.ground([('base', [])])

        atoms = sorted(list(self.ctl.theory_atoms), key=lambda atom: atom.term.name)
        self.assertEqual([str(atom) for atom in atoms], ['&a{}', '&b{}=1'])

        a, b = atoms
        self.assertTrue(a.literal >= 1)
        self.assertIsNone(a.guard)
        self.assertIsNotNone(b.guard)
        self.assertEqual(b.guard[0], "=")
        self.assertEqual(str(b.guard[1]), "1")
        self.assertEqual(len(a.elements), 0)

        self.assertEqual(a, a)
        self.assertNotEqual(a, b)
        self.assertNotEqual(hash(a), hash(b))
        self.assertNotEqual(a < b, b < a)

        self.assertRegex(repr(a), 'TheoryAtom(.*)')



'''
Functions and classes related to solving.

Examples
--------

The following example shows how to intercept models with a callback:

    >>> from clingo import Control
    >>>
    >>> ctl = Control(["0"])
    >>> ctl.add("base", [], "1 { a; b } 1.")
    >>> ctl.ground([("base", [])])
    >>> print(ctl.solve(on_model=print))
    Answer: a
    Answer: b
    SAT

The following example shows how to yield models:

    >>> from clingo import Control
    >>>
    >>> ctl = Control(["0"])
    >>> ctl.add("base", [], "1 { a; b } 1.")
    >>> ctl.ground([("base", [])])
    >>> with ctl.solve(yield_=True) as hnd:
    ...     for m in hnd:
    ...         print(m)
    ...     print(hnd.get())
    ...
    Answer: a
    Answer: b
    SAT

The following example shows how to solve asynchronously:

    >>> from clingo import Control
    >>>
    >>> ctl = Control(["0"])
    >>> ctl = clingo.Control("0")
    >>> ctl.add("base", [], "1 { a; b } 1.")
    >>> ctl.ground([("base", [])])
    >>> with ctl.solve(on_model=print, async_=True) as hnd:
    ...     # some computation here
    ...     hnd.wait():
    ...     print(hnd.get())
    ...
    Answer: a
    Answer: b
    SAT

This example shows how to solve both iteratively and asynchronously:

    >>> from clingo import Control
    >>>
    >>> ctl = Control(["0"])
    >>> ctl.configuration.solve.models = 0
    >>> ctl.add("base", [], "1 { a; b } 1.")
    >>> ctl.ground([("base", [])])
    >>> with ctl.solve(yield_=True, async_=True) as hnd:
    ...     while True:
    ...         hnd.resume()
    ...         # some computation here
    ...         _ = hnd.wait()
    ...         m = hnd.model()
    ...         if m is None:
    ...             print(hnd.get())
    ...             break
    ...         print(m)
    b
    a
    a b
    None
'''

from typing import ContextManager, Iterator, List, Optional, Sequence, Tuple, Union
from enum import Enum

from ._internal import _c_call, _c_call2, _ffi, _handle_error, _lib
from .util import Slice, SlicedSequence
from .symbol import Symbol
from .symbolic_atoms import SymbolicAtoms

__all__ = [ 'Model', 'ModelType', 'SolveControl', 'SolveHandle', 'SolveResult' ]

class SolveResult:
    '''
    Captures the result of a solve call.
    '''
    def __init__(self, rep):
        self._rep = rep

    @property
    def exhausted(self) -> bool:
        '''
        Determine if the search space was exhausted.
        '''
        return (_lib.clingo_solve_result_exhausted & self._rep) != 0

    @property
    def interrupted(self) -> bool:
        '''
        Determine if the search was interrupted.
        '''
        return (_lib.clingo_solve_result_interrupted & self._rep) != 0

    @property
    def satisfiable(self) -> Optional[bool]:
        '''
        `True` if the problem is satisfiable, `False` if the problem is
        unsatisfiable, or `None` if the satisfiablity is not known.
        '''
        if (_lib.clingo_solve_result_satisfiable & self._rep) != 0:
            return True
        if (_lib.clingo_solve_result_unsatisfiable & self._rep) != 0:
            return False
        return None

    @property
    def unknown(self) -> bool:
        '''
        Determine if the satisfiablity is not known.

        This is equivalent to satisfiable is None.
        '''
        return self.satisfiable is None

    @property
    def unsatisfiable(self) -> Optional[bool]:
        '''
        `True` if the problem is unsatisfiable, `False` if the problem is
        satisfiable, or `None` if the satisfiablity is not known.
        '''
        if (_lib.clingo_solve_result_unsatisfiable & self._rep) != 0:
            return True
        if (_lib.clingo_solve_result_satisfiable & self._rep) != 0:
            return False
        return None

    def __str__(self):
        if self.satisfiable:
            return "SAT"
        if self.unsatisfiable:
            return "UNSAT"
        return "UNKNOWN"

    def __repr__(self):
        return f"SolveResult({self._rep})"

class SolveControl:
    '''
    Object that allows for controlling a running search.
    '''
    def __init__(self, rep):
        self._rep = rep

    def add_clause(self, literals: Sequence[Union[Tuple[Symbol,bool],int]]) -> None:
        '''
        Add a clause that applies to the current solving step during the search.

        Parameters
        ----------
        literals
            List of literals either represented as pairs of symbolic atoms and
            Booleans or as program literals.

        Notes
        -----
        This function can only be called in a model callback or while iterating
        when using a `SolveHandle`.
        '''
        atoms = self.symbolic_atoms
        p_lits = _ffi.new('clingo_literal_t[]', len(literals))
        for i, lit in enumerate(literals):
            if isinstance(lit, int):
                p_lits[i] = lit
            else:
                atom = atoms[lit[0]]
                if atom is not None:
                    slit = atom.literal
                else:
                    slit = -1
                p_lits[i] = slit if lit[1] else -slit

        _handle_error(_lib.clingo_solve_control_add_clause(self._rep, p_lits, len(literals)))

    def _invert(self, lit: Union[Tuple[Symbol,bool],int]) -> Union[Tuple[Symbol,bool],int]:
        if isinstance(lit, int):
            return -lit
        return lit[0], not lit[1]

    def add_nogood(self, literals: Sequence[Union[Tuple[Symbol,bool],int]]) -> None:
        '''
        Equivalent to `SolveControl.add_clause` with the literals inverted.
        '''
        self.add_clause([self._invert(lit) for lit in literals])

    @property
    def symbolic_atoms(self) -> SymbolicAtoms:
        '''
        `clingo.symbolic_atoms.SymbolicAtoms` object to inspect the symbolic atoms.
        '''
        atoms = _c_call('clingo_symbolic_atoms_t*', _lib.clingo_solve_control_symbolic_atoms, self._rep)
        return SymbolicAtoms(atoms)

class ModelType(Enum):
    '''
    Enumeration of the different types of models.
    '''
    BraveConsequences = _lib.clingo_model_type_brave_consequences
    '''
    The model stores the set of brave consequences.
    '''
    CautiousConsequences = _lib.clingo_model_type_cautious_consequences
    '''
    The model stores the set of cautious consequences.
    '''
    StableModel = _lib.clingo_model_type_stable_model
    '''
    The model captures a stable model.
    '''

class _SymbolSequence(Sequence[Symbol]):
    '''
    Helper class to efficiently store sequences of symbols.
    '''
    def __init__(self, p_symbols):
        self._p_symbols = p_symbols

    def __len__(self):
        return len(self._p_symbols)

    def __getitem__(self, slc):
        if isinstance(slc, slice):
            return SlicedSequence(self, Slice(slc))
        if slc < 0:
            slc += len(self)
        if slc < 0 or slc >= len(self):
            raise IndexError('invalid index')
        return Symbol(self._p_symbols[slc])

    def __iter__(self):
        for i in range(len(self)):
            yield Symbol(self._p_symbols[i])

    def __str__(self):
        return f'[{", ".join(str(sym) for sym in self)}]'

    def __repr__(self):
        return f'[{", ".join(repr(sym) for sym in self)}]'

class Model:
    '''
    Provides access to a model during a solve call and provides a
    `SolveContext` object to influence the running search.

    Notes
    -----
    The string representation of a model object is similar to the output of
    models by clingo using the default output.

    `Model` objects cannot be constructed from Python. Instead they are obained
    during solving (see `Control.solve`). Furthermore, the lifetime of a model
    object is limited to the scope of the callback it was passed to or until
    the search for the next model is started. They must not be stored for later
    use.
    '''
    def __init__(self, rep):
        self._rep = rep

    def contains(self, atom: Symbol) -> bool:
        '''
        Efficiently check if an atom is contained in the model.

        Parameters
        ----------
        atom
            The atom to lookup.

        Returns
        -------
        Whether the given atom is contained in the model.

        Notes
        -----
        The atom must be represented using a function symbol.
        '''
        # pylint: disable=protected-access
        return _c_call('bool', _lib.clingo_model_contains, self._rep, atom._rep)

    def extend(self, symbols: Sequence[Symbol]) -> None:
        '''
        Extend a model with the given symbols.

        Parameters
        ----------
        symbols
            The symbols to add to the model.

        Notes
        -----
        This only has an effect if there is an underlying clingo application,
        which will print the added symbols.
        '''
        # pylint: disable=protected-access
        c_symbols = _ffi.new('clingo_symbol_t[]', len(symbols))
        for i, sym in enumerate(symbols):
            c_symbols[i] = sym._rep
        _handle_error(_lib.clingo_model_extend(self._rep, c_symbols, len(symbols)))

    def is_true(self, literal: int) -> bool:
        '''
        Check if the given program literal is true.

        Parameters
        ----------
        literal
            The given program literal.

        Returns
        -------
        Whether the given program literal is true.
        '''
        return _c_call('bool', _lib.clingo_model_is_true, self._rep, literal)

    def symbols(self, atoms: bool=False, terms: bool=False, shown: bool=False, csp: bool=False,
                theory: bool=False, complement: bool=False) -> Sequence[Symbol]:
        '''
        Return the list of atoms, terms, or CSP assignments in the model.

        Parameters
        ----------
        atoms
            Select all atoms in the model (independent of `#show` statements).
        terms
            Select all terms displayed with `#show` statements in the model.
        shown
            Select all atoms and terms as outputted by clingo.
        csp
            Select all csp assignments (independent of `#show` statements).
        theory
            Select atoms added with `Model.extend`.
        complement
            Return the complement of the answer set w.r.t. to the atoms known
            to the grounder. (Does not affect csp assignments.)

        Returns
        -------
        The selected symbols.

        Notes
        -----
        Atoms are represented using functions (`Symbol` objects), and CSP
        assignments are represented using functions with name `"$"` where the
        first argument is the name of the CSP variable and the second its
        value.
        '''
        show = 0
        if atoms:
            show |= _lib.clingo_show_type_atoms
        if terms:
            show |= _lib.clingo_show_type_terms
        if shown:
            show |= _lib.clingo_show_type_shown
        if csp:
            show |= _lib.clingo_show_type_csp
        if theory:
            show |= _lib.clingo_show_type_theory
        if complement:
            show |= _lib.clingo_show_type_complement

        size = _c_call('size_t', _lib.clingo_model_symbols_size, self._rep, show)

        p_symbols = _ffi.new('clingo_symbol_t[]', size)
        _handle_error(_lib.clingo_model_symbols(self._rep, show, p_symbols, size))

        return _SymbolSequence(p_symbols)

    def __str__(self):
        return " ".join(map(str, self.symbols(shown=True)))

    def __repr__(self):
        return f'Model({self._rep!r})'

    @property
    def context(self) -> SolveControl:
        '''
        Object that allows for controlling the running search.
        '''
        ctl = _c_call('clingo_solve_control_t*', _lib.clingo_model_context, self._rep)
        return SolveControl(ctl)

    @property
    def cost(self) -> List[int]:
        '''
        Return the list of integer cost values of the model.

        The return values correspond to clasp's cost output.
        '''
        size = _c_call('size_t', _lib.clingo_model_cost_size, self._rep)

        p_costs = _ffi.new('int64_t[]', size)
        _handle_error(_lib.clingo_model_cost(self._rep, p_costs, size))

        return list(p_costs)

    @property
    def number(self) -> int:
        '''
        The running number of the model.
        '''
        return _c_call('uint64_t', _lib.clingo_model_number, self._rep)

    @property
    def optimality_proven(self) -> bool:
        '''
        Whether the optimality of the model has been proven.
        '''
        return _c_call('bool', _lib.clingo_model_optimality_proven, self._rep)

    @property
    def thread_id(self) -> int:
        '''
        The id of the thread which found the model.
        '''
        return _c_call('clingo_id_t', _lib.clingo_model_thread_id, self._rep)

    @property
    def type(self) -> ModelType:
        '''
        The type of the model.
        '''
        return ModelType(_c_call('clingo_model_type_t', _lib.clingo_model_type, self._rep))

class SolveHandle(ContextManager['SolveHandle']):
    '''
    Handle for solve calls.

    They can be used to control solving, like, retrieving models or cancelling
    a search.

    See Also
    --------
    Control.solve

    Notes
    -----
    A `SolveHandle` is a context manager and must be used with Python's `with`
    statement.

    Blocking functions in this object release the GIL. They are not thread-safe
    though.
    '''
    def __init__(self, rep, handler):
        self._rep = rep
        self._handler = handler

    def __iter__(self) -> Iterator[Model]:
        while True:
            self.resume()
            m = self.model()
            if m is None:
                break
            yield m

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        _handle_error(_lib.clingo_solve_handle_close(self._rep), self._handler)
        return False

    def cancel(self) -> None:
        '''
        Cancel the running search.

        See Also
        --------
        clingo.control.Control.interrupt
        '''
        _handle_error(_lib.clingo_solve_handle_cancel(self._rep), self._handler)

    def core(self) -> List[int]:
        '''
        The subset of assumptions that made the problem unsatisfiable.
        '''
        core, size = _c_call2('clingo_literal_t*', 'size_t', _lib.clingo_solve_handle_core,
                              self._rep, handler=self._handler)
        return [core[i] for i in range(size)]

    def get(self) -> SolveResult:
        '''
        Get the result of a solve call.

        If the search is not completed yet, the function blocks until the
        result is ready.
        '''
        res = _c_call('clingo_solve_result_bitset_t', _lib.clingo_solve_handle_get, self._rep, handler=self._handler)
        return SolveResult(res)

    def model(self) -> Optional[Model]:
        '''
        Get the current model if there is any.
        '''
        p_model = _ffi.new('clingo_model_t**')
        _handle_error(
            _lib.clingo_solve_handle_model(self._rep, p_model),
            self._handler)
        if p_model[0] == _ffi.NULL:
            return None
        return Model(p_model[0])

    def resume(self) -> None:
        '''
        Discards the last model and starts searching for the next one.

        Notes
        -----
        If the search has been started asynchronously, this function starts the
        search in the background.
        '''
        _handle_error(_lib.clingo_solve_handle_resume(self._rep), self._handler)

    def wait(self, timeout: Optional[float]=None) -> bool:
        '''
        Wait for solve call to finish or the next result with an optional timeout.

        If a timeout is given, the behavior of the function changes depending
        on the sign of the timeout. If a postive timeout is given, the function
        blocks for the given amount time or until a result is ready. If the
        timeout is negative, the function will block until a result is ready,
        which also corresponds to the behavior of the function if no timeout is
        given. A timeout of zero can be used to poll if a result is ready.

        Parameters
        ----------
        timeout
            If a timeout is given, the function blocks for at most timeout seconds.

        Returns
        -------
        Indicates whether the solve call has finished or the next result is ready.
        '''
        p_res = _ffi.new('bool*')
        _lib.clingo_solve_handle_wait(self._rep, -1 if timeout is None else timeout, p_res)
        return p_res[0]



import sys
import site
from textwrap import dedent
from skbuild import setup

if not site.ENABLE_USER_SITE and "--user" in sys.argv[1:]:
    site.ENABLE_USER_SITE = True

setup(
    version = '5.5.2',
    name = 'clingo',
    description = 'CFFI-based bindings to the clingo solver.',
    long_description = dedent('''\
        This package provides CFFI-based bindings to the clingo solver.

        Clingo is part of the [Potassco](https://potassco.org) project for *Answer Set Programming* (ASP).
        ASP offers a simple and powerful modeling language to describe combinatorial problems as *logic programs*.
        The *clingo* system then takes such a logic program and computes *answer sets* representing solutions to the given problem.
        To get an idea, check our [Getting Started](https://potassco.org/doc/start/) page and the [online version](https://potassco.org/clingo/run/) of clingo.

        Please check the the [API documentation](https://potassco.org/clingo/python-api/current/clingo/) on how to use this module.
        '''),
    long_description_content_type='text/markdown',
    author = 'Roland Kaminski',
    author_email = 'kaminski@cs.uni-potsdam.de',
    license = 'MIT',
    url = 'https://github.com/potassco/clingo',
    install_requires=[ 'cffi' ],
    cmake_args=[ '-DCLINGO_MANAGE_RPATH=OFF',
                 '-DCLINGO_BUILD_APPS=OFF',
                 '-DCLINGO_BUILD_WITH_PYTHON=pip',
                 '-DCLINGO_BUILD_WITH_LUA=OFF',
                 '-DPYCLINGO_INSTALL_DIR=libpyclingo' ],
    packages=[ 'clingo' ],
    package_data={ 'clingo': [ 'py.typed', 'import__clingo.lib', 'clingo.h', 'clingo.hh' ] },
    package_dir={ '': 'libpyclingo' },
    python_requires=">=3.6"
)



'''
Tests for configuration and statistics.
'''

from unittest import TestCase
from typing import cast
from clingo import Configuration, Control

class TestConfig(TestCase):
    '''
    Tests for configuration and statistics.
    '''
    def test_config(self):
        '''
        Test configuration.
        '''
        ctl = Control(['-t', '2'])
        self.assertIn('solver', ctl.configuration.keys)
        self.assertEqual(len(ctl.configuration.solver), 2)
        self.assertIsInstance(ctl.configuration.solver[0], Configuration)
        conf = cast(Configuration, ctl.configuration.solver[0])
        self.assertIsInstance(conf.heuristic, str)
        self.assertIsInstance(conf.description('heuristic'), str)
        conf.heuristic = 'berkmin'
        self.assertTrue(conf.heuristic.startswith('berkmin'))

    def test_simple_stats(self):
        '''
        Test simple statistics.
        '''
        ctl = Control(['-t', '2', '--stats=2'])
        ctl.add('base', [], '1 { a; b }.')
        ctl.ground([('base', [])])
        ctl.solve()
        stats = ctl.statistics
        self.assertGreaterEqual(stats['problem']['lp']['atoms'], 2)
        self.assertGreaterEqual(stats['solving']['solvers']['choices'], 1)

    def test_user_stats(self):
        '''
        Test user statistics.
        '''
        def on_statistics(step, accu):
            step['test'] = {'a': 0, 'b': [1, 2], 'c': {'d': 3}}
            accu['test'] = step['test']
            step['test'] = {'a': lambda a: a+1,
                            'e': lambda a: 4 if a is None else 0,
                            'b': [-1, 2, 3]}
            self.assertEqual(len(step['test']), 4)
            self.assertEqual(len(step['test']['b']), 3)
            self.assertEqual(len(step['test']['c']), 1)
            self.assertIn('a', step['test'])
            self.assertEqual(sorted(step['test']), ['a', 'b', 'c', 'e'])
            self.assertEqual(sorted(step['test'].keys()), ['a', 'b', 'c', 'e'])
            self.assertEqual(sorted(step['test']['c'].items()), [('d', 3.0)])
            self.assertEqual(sorted(step['test']['c'].values()), [3.0])

            step['test']['b'][1] = 99
            self.assertEqual(step['test']['b'][1], 99)
            step['test']['b'].extend([3, 4])
            step['test']['b'] += [3, 4]

        ctl = Control(['-t', '2', '--stats=2'])
        ctl.add('base', [], '1 { a; b }.')
        ctl.ground([('base', [])])
        ctl.solve(on_statistics=on_statistics)
        stats = ctl.statistics
        self.assertEqual(stats['user_step']['test'], {'a': 1.0,
                                                      'b': [-1.0, 99.0, 3.0, 3.0, 4.0, 3.0, 4.0],
                                                      'c': {'d': 3.0},
                                                      'e': 4.0})
        self.assertEqual(stats['user_accu']['test'], {'a': 0, 'b': [1, 2], 'c': {'d': 3}})



'''
Tests for the propagator.
'''

from unittest import TestCase
from typing import cast
from clingo import (Assignment, Control, Function, PropagateControl, PropagateInit, Propagator, PropagatorCheckMode,
                    SolveResult, SymbolicAtom)

from .util import _MCB, _check_sat, _p

class TestPropagatorControl(Propagator):
    '''
    Test functions in PropagateControl.
    '''
    def __init__(self, case: TestCase):
        self._case = case
        self._lit_a = 0

    def init(self, init: PropagateInit):
        init.check_mode = PropagatorCheckMode.Off
        self._case.assertEqual(init.check_mode, PropagatorCheckMode.Off)
        self._case.assertEqual(init.number_of_threads, 1)
        a = init.symbolic_atoms[Function('a')]
        self._case.assertIsNotNone(a)
        self._lit_a = init.solver_literal(cast(SymbolicAtom, a).literal)
        init.add_watch(-self._lit_a)

    def propagate(self, control: PropagateControl, changes):
        ass = control.assignment
        trail = ass.trail
        lvl = ass.decision_level
        self._case.assertIn(-self._lit_a, changes)
        self._case.assertGreaterEqual(lvl, 1)
        self._case.assertGreaterEqual(ass.level(self._lit_a), 1)
        self._case.assertGreaterEqual(len(trail), 1)
        self._case.assertGreaterEqual(len(list(trail)), 1)
        self._case.assertEqual(trail[trail.begin(lvl)], -self._lit_a)
        self._case.assertEqual(list(trail[trail.begin(lvl):trail.end(lvl)]), [-self._lit_a])
        self._case.assertEqual(ass.decision(lvl), -self._lit_a)
        self._case.assertEqual(control.thread_id, 0)
        self._case.assertTrue(control.has_watch(-self._lit_a))
        self._case.assertTrue(control.propagate())
        self._case.assertFalse(control.add_clause([self._lit_a]))

    def undo(self, thread_id, assignment, changes):
        self._case.assertEqual(thread_id, 0)
        self._case.assertIn(-self._lit_a, changes)

class TestPropagatorInit(Propagator):
    '''
    Test functions in PropagateInit.
    '''
    def __init__(self, case: TestCase):
        self._case = case

    def init(self, init: PropagateInit):
        a = init.symbolic_atoms[Function('a')]
        b = init.symbolic_atoms[Function('b')]
        c = init.symbolic_atoms[Function('c')]
        self._case.assertIsNotNone(a)
        self._case.assertIsNotNone(b)
        self._case.assertIsNotNone(c)
        lit_a = init.solver_literal(cast(SymbolicAtom, a).literal)
        lit_b = init.solver_literal(cast(SymbolicAtom, b).literal)
        lit_c = init.solver_literal(cast(SymbolicAtom, c).literal)
        lit = init.add_literal()
        # a <=> b
        init.add_clause([lit_a, -lit])
        init.add_clause([lit, -lit_a])
        init.add_clause([lit_b, -lit])
        init.add_clause([lit, -lit_b])
        # c <=> {a, b} >= 2
        init.add_weight_constraint(lit_c, [(lit_a, 1), (lit_b, 1)], 2, 0)
        init.add_minimize(lit_a, -1)
        self._case.assertTrue(init.propagate())
        self._case.assertEqual(list(init.theory_atoms), [])
        # test assignment
        self._case.assertIsNone(init.assignment.value(lit_a))
        self._case.assertFalse(init.assignment.is_true(lit_a))
        self._case.assertFalse(init.assignment.is_false(lit_a))
        self._case.assertFalse(init.assignment.is_fixed(lit_a))
        self._case.assertEqual(init.assignment.decision_level, 0)
        self._case.assertTrue(init.assignment.has_literal(lit_a))
        self._case.assertFalse(init.assignment.has_conflict)
        self._case.assertFalse(init.assignment.is_total)
        self._case.assertEqual(init.assignment.root_level, 0)
        self._case.assertGreaterEqual(len(init.assignment), 4)
        self._case.assertGreaterEqual(len(list(init.assignment)), 4)
        self._case.assertEqual(len(init.assignment[0:4][0:4:2]), 2)
        self._case.assertEqual(init.assignment[0:4][0:4:2][1], init.assignment[2])

class TestPropagator(Propagator):
    '''
    Test adding literals while solving.
    '''
    def __init__(self, case: TestCase):
        self._case = case
        self._added = False

    def check(self, control: PropagateControl):
        if not self._added:
            self._added = True
            lit = control.add_literal()
            self._case.assertFalse(control.has_watch(lit))
            control.add_watch(lit)
            self._case.assertTrue(control.has_watch(lit))
            control.remove_watch(lit)
            self._case.assertFalse(control.has_watch(lit))

class TestHeuristic(Propagator):
    '''
    Test decide.
    '''
    def __init__(self, case: TestCase):
        self._case = case
        self._lit_a = 0
        self._lit_b = 0

    def init(self, init: PropagateInit):
        a = init.symbolic_atoms[Function('a')]
        b = init.symbolic_atoms[Function('b')]
        self._case.assertIsNotNone(a)
        self._case.assertIsNotNone(b)
        self._lit_a = init.solver_literal(cast(SymbolicAtom, a).literal)
        self._lit_b = init.solver_literal(cast(SymbolicAtom, b).literal)

    def decide(self, thread_id: int, assignment: Assignment, fallback: int) -> int:
        self._case.assertEqual(thread_id, 0)
        if assignment.is_free(self._lit_a):
            return self._lit_a
        if assignment.is_free(self._lit_b):
            return -self._lit_b
        return fallback

class TestSymbol(TestCase):
    '''
    Tests basic solving and related functions.
    '''

    def setUp(self):
        self.mcb = _MCB()
        self.ctl = Control(['0'])

    def tearDown(self):
        self.mcb = None
        self.ctl = None

    def test_propagator_control(self):
        '''
        Test PropagateControl.
        '''
        self.ctl.add("base", [], "{a}.")
        self.ctl.ground([("base", [])])
        self.ctl.register_propagator(TestPropagatorControl(self))
        _check_sat(self, cast(SolveResult, self.ctl.solve(on_model=self.mcb.on_model, yield_=False, async_=False)))
        self.assertEqual(self.mcb.models, _p(['a']))

    def test_propagator_init(self):
        '''
        Test PropagateInit and Assignment.
        '''
        self.ctl.add("base", [], "{a; b; c}.")
        self.ctl.ground([("base", [])])
        self.ctl.register_propagator(TestPropagatorInit(self))
        _check_sat(self, cast(SolveResult, self.ctl.solve(on_model=self.mcb.on_model, yield_=False, async_=False)))
        self.assertEqual(self.mcb.models[-1:], _p(['a', 'b', 'c']))

    def test_propagator(self):
        '''
        Test adding literals while solving.
        '''
        self.ctl.add("base", [], "")
        self.ctl.ground([("base", [])])
        self.ctl.register_propagator(TestPropagator(self))
        _check_sat(self, cast(SolveResult, self.ctl.solve(on_model=self.mcb.on_model, yield_=False, async_=False)))
        self.assertEqual(self.mcb.models, _p([], []))

    def test_heurisitc(self):
        '''
        Test decide.
        '''
        self.ctl = Control(['1'])
        self.ctl.add("base", [], "{a;b}.")
        self.ctl.ground([("base", [])])
        self.ctl.register_propagator(TestHeuristic(self))
        self.ctl.solve(on_model=self.mcb.on_model)
        self.assertEqual(self.mcb.models, _p(['a']))



'''
Functions and classes to work with theory atoms.

Examples
--------
    >>> from clingo.control import Control
    >>>
    >>> ctl = Control()
    >>> ctl.add('base', [], """\\
    ... #theory example {
    ...     t { };
    ...     &a/0 : t, head
    ... }.
    ... {c}.
    ... &a { t: c }.
    ... """)
    >>> ctl.ground([('base', [])])
    >>> atm = next(ctl.theory_atoms)
    >>> print(atm)
    &a{t: c}
    >>> elm = atm.elements[0]
    >>> print(elm)
    t: c
'''

from typing import List, Optional, Tuple
from enum import Enum
from functools import total_ordering

from ._internal import _c_call, _c_call2, _lib, _str, _to_str

__all__ = [ 'TheoryAtom', 'TheoryElement', 'TheoryTerm', 'TheoryTermType' ]

class TheoryTermType(Enum):
    '''
    Enumeration of theory term types.
    '''
    Function = _lib.clingo_theory_term_type_function
    '''
    For a function theory terms.
    '''
    List = _lib.clingo_theory_term_type_list
    '''
    For list theory terms.
    '''
    Number = _lib.clingo_theory_term_type_number
    '''
    For numeric theory terms.
    '''
    Set = _lib.clingo_theory_term_type_set
    '''
    For set theory terms.
    '''
    Symbol = _lib.clingo_theory_term_type_symbol
    '''
    For symbolic theory terms (symbol here means the term is a string).
    '''
    Tuple = _lib.clingo_theory_term_type_tuple
    '''
    For tuple theory terms.
    '''

@total_ordering
class TheoryTerm:
    '''
    `TheoryTerm` objects represent theory terms.

    Theory terms have a readable string representation, implement Python's rich
    comparison operators, and can be used as dictionary keys.
    '''
    def __init__(self, rep, idx):
        self._rep = rep
        self._idx = idx

    def __hash__(self):
        return self._idx

    def __eq__(self, other):
        return self._idx == other._idx

    def __lt__(self, other):
        return self._idx < other._idx

    def __str__(self):
        return _str(_lib.clingo_theory_atoms_term_to_string_size,
                    _lib.clingo_theory_atoms_term_to_string,
                    self._rep, self._idx)

    def __repr__(self):
        return f'TheoryTerm({self._rep!r})'

    @property
    def arguments(self) -> List['TheoryTerm']:
        '''
        The arguments of the term (for functions, tuples, list, and sets).
        '''
        args, size = _c_call2('clingo_id_t*', 'size_t', _lib.clingo_theory_atoms_term_arguments, self._rep, self._idx)
        return [TheoryTerm(self._rep, args[i]) for i in range(size)]

    @property
    def name(self) -> str:
        '''
        The name of the term (for symbols and functions).
        '''
        return _to_str(_c_call('char*', _lib.clingo_theory_atoms_term_name, self._rep, self._idx))

    @property
    def number(self) -> int:
        '''
        The numeric representation of the term (for numbers).
        '''
        return _c_call('int', _lib.clingo_theory_atoms_term_number, self._rep, self._idx)

    @property
    def type(self) -> TheoryTermType:
        '''
        The type of the theory term.
        '''
        type_ = _c_call('clingo_theory_term_type_t', _lib.clingo_theory_atoms_term_type, self._rep, self._idx)
        return TheoryTermType(type_)

@total_ordering
class TheoryElement:
    '''
    Class to represent theory elements.

    Theory elements have a readable string representation, implement Python's rich
    comparison operators, and can be used as dictionary keys.
    '''
    def __init__(self, rep, idx):
        self._rep = rep
        self._idx = idx

    def __hash__(self):
        return self._idx

    def __eq__(self, other):
        return self._idx == other._idx

    def __lt__(self, other):
        return self._idx < other._idx

    def __str__(self):
        return _str(_lib.clingo_theory_atoms_element_to_string_size,
                    _lib.clingo_theory_atoms_element_to_string,
                    self._rep, self._idx)

    def __repr__(self):
        return f'TheoryElement({self._rep!r})'

    @property
    def condition(self) -> List[int]:
        '''
        The condition of the element in form of a list of program literals.
        '''
        cond, size = _c_call2('clingo_literal_t*', 'size_t', _lib.clingo_theory_atoms_element_condition,
                              self._rep, self._idx)
        return [cond[i] for i in range(size)]

    @property
    def condition_id(self) -> int:
        '''
        Each condition has an id, which is a temporary program literal. This id
        can be passed to `clingo.propagator.PropagateInit.solver_literal` to
        obtain a corresponding solver literal.
        '''
        return _c_call('clingo_literal_t', _lib.clingo_theory_atoms_element_condition_id, self._rep, self._idx)

    @property
    def terms(self) -> List[TheoryTerm]:
        '''
        The tuple of the element.
        '''
        terms, size = _c_call2('clingo_id_t*', 'size_t', _lib.clingo_theory_atoms_element_tuple, self._rep, self._idx)
        return [TheoryTerm(self._rep, terms[i]) for i in range(size)]

@total_ordering
class TheoryAtom:
    '''
    Class to represent theory atoms.

    Theory atoms have a readable string representation, implement Python's rich
    comparison operators, and can be used as dictionary keys.
    '''
    def __init__(self, rep, idx):
        self._rep = rep
        self._idx = idx

    def __hash__(self):
        return self._idx

    def __eq__(self, other):
        return self._idx == other._idx

    def __lt__(self, other):
        return self._idx < other._idx

    def __str__(self):
        return _str(_lib.clingo_theory_atoms_atom_to_string_size,
                    _lib.clingo_theory_atoms_atom_to_string,
                    self._rep, self._idx)

    def __repr__(self):
        return f'TheoryAtom({self._rep!r})'

    @property
    def elements(self) -> List[TheoryElement]:
        '''
        The elements of the atom.
        '''
        elems, size = _c_call2('clingo_id_t*', 'size_t', _lib.clingo_theory_atoms_atom_elements, self._rep, self._idx)
        return [TheoryElement(self._rep, elems[i]) for i in range(size)]

    @property
    def guard(self) -> Optional[Tuple[str, TheoryTerm]]:
        '''
        The guard of the atom or None if the atom has no guard.
        '''
        if not _c_call('bool', _lib.clingo_theory_atoms_atom_has_guard, self._rep, self._idx):
            return None

        conn, term = _c_call2('char*', 'clingo_id_t', _lib.clingo_theory_atoms_atom_guard, self._rep, self._idx)

        return (_to_str(conn), TheoryTerm(self._rep, term))

    @property
    def literal(self) -> int:
        '''
        The program literal associated with the atom.
        '''
        return _c_call('clingo_literal_t', _lib.clingo_theory_atoms_atom_literal, self._rep, self._idx)

    @property
    def term(self) -> TheoryTerm:
        '''
        The term of the atom.
        '''
        term = _c_call('clingo_id_t', _lib.clingo_theory_atoms_atom_term, self._rep, self._idx)
        return TheoryTerm(self._rep, term)
